import functools
import typing
from xoa_driver.internals.commands import (
    P_CAPABILITIES,
    P_SPEED,
    P_SPEEDREDUCTION,
    P_INTERFRAMEGAP,
    P_MACADDRESS,
    P_IPADDRESS,
    P_ARPREPLY,
    P_PINGREPLY,
    P_PAUSE,
    P_RANDOMSEED,
    P_LOOPBACK,
    P_TRAFFIC,
    # P_CAPTURE,
    P_XMITONE,
    P_LATENCYOFFSET,
    P_LATENCYMODE,
    P_AUTOTRAIN,
    P_MIXWEIGHTS,
    P_TRAFFICERR,
    P_GAPMONITOR,
    P_CHECKSUM,
    P_MIXLENGTH,
    P_ARPRXTABLE,
    P_NDPRXTABLE,
    P_MULTICAST,
    P_MULTICASTEXT,
    P_MCSRCLIST,
    P_MULTICASTHDR,
    P_TXMODE,
    P_RATEFRACTION,
    P_RATEPPS,
    P_RATEL2BPS,
    P_PAYLOADMODE,
    P_TXENABLE,
    P_MAXHEADERLENGTH,
    P_TXTIMELIMIT,
    P_TXTIME,
    P_XMITONETIME,
    P_IPV6ADDRESS,
    P_ARPV6REPLY,
    P_PINGV6REPLY,
    P_ERRORS,
    P_TXPREPARE,
    P_TXDELAY,
    P_TPLDMODE,
    P_TXPACKETLIMIT,
    P_PFCENABLE,
    P_TXBURSTPERIOD,
    P_CAPABILITIES_EXT,
)
if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.hli_v1.ports import base_port
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as idx_mgr
from xoa_driver.internals.state_storage import ports_state
from xoa_driver.internals.hli_v1.indices.length_term import LengthTermIdx
from xoa_driver.internals.hli_v1.indices.match_term import MatchTermIdx

from .port_capture import PortCapture

LengthTermIndices = idx_mgr.IndexManager[LengthTermIdx]
MatchTermIndices = idx_mgr.IndexManager[MatchTermIdx]


class TxSinglePacket:
    """L23 port single-packet transmission"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.send = P_XMITONE(conn, module_id, port_id)
        """Send one packet from the L23 port without a stream config.

        :type: P_XMITONE
        """

        self.time = P_XMITONETIME(conn, module_id, port_id)
        """The time at which the latest packet was transmitted using the P_XMITONE` command.

        :type: P_XMITONETIME
        """


class TxConfiguration:
    """L23 port TX configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mode = P_TXMODE(conn, module_id, port_id)
        """L23 port TX mode.

        :type: P_TXMODE
        """

        self.enable = P_TXENABLE(conn, module_id, port_id)
        """Enabling L23 port TX.

        :type: P_TXENABLE
        """

        self.time_limit = P_TXTIMELIMIT(conn, module_id, port_id)
        """L23 port TX time limit.

        :type: P_TXTIMELIMIT
        """

        self.time = P_TXTIME(conn, module_id, port_id)
        """L23 port TX time.

        :type: P_TXTIME
        """

        self.prepare = P_TXPREPARE(conn, module_id, port_id)
        """Prepare L23 port for transmission.

        :type: P_TXPREPARE
        """

        self.delay = P_TXDELAY(conn, module_id, port_id)
        """L23 port TX delay.

        :type: P_TXDELAY
        """

        self.packet_limit = P_TXPACKETLIMIT(conn, module_id, port_id)
        """L23 port TX packet limit

        :type: P_TXPACKETLIMIT
        """

        self.burst_period = P_TXBURSTPERIOD(conn, module_id, port_id)
        """L23 port TX burst period.

        :type: P_TXBURSTPERIOD
        """


class Rate:
    """L23 port TX rate"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.fraction = P_RATEFRACTION(conn, module_id, port_id)
        """L23 port rate in ppm.

        :type: P_RATEFRACTION
        """

        self.pps = P_RATEPPS(conn, module_id, port_id)
        """L23 port rate in packets per second.

        :type: P_RATEPPS
        """

        self.l2_bps = P_RATEL2BPS(conn, module_id, port_id)
        """L23 port rate in L2 bits per second.

        :type: P_RATEL2BPS
        """


class Multicast:
    """L23 port multicast configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mode = P_MULTICAST(conn, module_id, port_id)
        """L23 port multicast mode.

        :type: P_MULTICAST
        """

        self.mode_extended = P_MULTICASTEXT(conn, module_id, port_id)
        """L23 port multicast extended mode.

        :type: P_MULTICASTEXT
        """

        self.source_list = P_MCSRCLIST(conn, module_id, port_id)
        """L23 port multicast source list.

        :type: P_MCSRCLIST
        """

        self.header = P_MULTICASTHDR(conn, module_id, port_id)
        """L23 port multicast IGMP header.

        :type: P_MULTICASTHDR
        """


class IPv4:
    """L23 port IPv4 configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.address = P_IPADDRESS(conn, module_id, port_id)
        """L23 port IPv4 address.

        :type: P_IPADDRESS
        """

        self.arp_reply = P_ARPREPLY(conn, module_id, port_id)
        """L23 port reply to ARP request.

        :type: P_ARPREPLY
        """

        self.ping_reply = P_PINGREPLY(conn, module_id, port_id)
        """L23 port reply to PING request.

        :type: P_PINGREPLY
        """


class IPv6:
    """L23 port IPv6 configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.address = P_IPV6ADDRESS(conn, module_id, port_id)
        """L23 port IPv6 address.

        :type: P_IPV6ADDRESS
        """

        self.arp_reply = P_ARPV6REPLY(conn, module_id, port_id)
        """L23 port reply to NDP Neighbor Solicitation.

        :type: P_ARPV6REPLY
        """

        self.ping_reply = P_PINGV6REPLY(conn, module_id, port_id)
        """L23 port reply to PINGv6 request.

        :type: P_PINGV6REPLY
        """


class NetworkConfiguration:  # will be extended in genuine ports
    """L23 port networking configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mac_address = P_MACADDRESS(conn, module_id, port_id)
        """L23 port MAC address.

        :type: P_MACADDRESS
        """

        self.ipv4 = IPv4(conn, module_id, port_id)
        """L23 port IPv4 address configuration.
        """

        self.ipv6 = IPv6(conn, module_id, port_id)
        """L23 port IPv6 address configuration.
        """


class LatencyConfiguration:
    """L23 port latency configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.offset = P_LATENCYOFFSET(conn, module_id, port_id)
        """L23 port latency offset.

        :type: P_LATENCYOFFSET
        """

        self.mode = P_LATENCYMODE(conn, module_id, port_id)
        """L23 port latency measurement mode.

        :type: P_LATENCYMODE
        """


class Mix:
    """L23 port IMIX configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.weights = P_MIXWEIGHTS(conn, module_id, port_id)
        """L23 port IMIX weights

        :type: P_MIXWEIGHTS
        """

        self.lengths = tuple(
            P_MIXLENGTH(conn, module_id, port_id, idx)
            for idx in range(16)
        )  # TODO: need to add manager for handle specific indices only
        """L23 port IMIX lengths.

        :type: P_MIXLENGTH
        """


class Speed:
    """L23 port speed configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.current = P_SPEED(conn, module_id, port_id)
        """L23 port current speed in units of Mbps.

        :type: P_SPEED
        """

        self.reduction = P_SPEEDREDUCTION(conn, module_id, port_id)
        """L23 port speed reduction in ppm.

        :type: P_SPEEDREDUCTION
        """


class Traffic:
    """L23 port traffic generation"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.state = P_TRAFFIC(conn, module_id, port_id)
        """L23 port traffic status and action.

        :type: P_TRAFFIC
        """

        self.error = P_TRAFFICERR(conn, module_id, port_id)
        """L23 port traffic error.

        :type: P_TRAFFICERR
        """


class BasePortL23(base_port.BasePort[ports_state.PortL23LocalState]):
    """L23 port layout which is relevant to all L23 ports."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.capabilities = P_CAPABILITIES(conn, module_id, port_id)
        """L23 port capabilities.

        :type: P_CAPABILITIES
        """

        self.capabilities_ext = P_CAPABILITIES_EXT(conn, module_id, port_id)
        """L23 port capabilities ext.

        :type: P_CAPABILITIES_EXT
        """

        self.pause = P_PAUSE(conn, module_id, port_id)
        """L23 port response to Ethernet PAUSE frames.

        :type: P_PAUSE
        """

        self.loop_back = P_LOOPBACK(conn, module_id, port_id)
        """L23 port loopback mode.

        :type: P_LOOPBACK
        """

        self.errors_count = P_ERRORS(conn, module_id, port_id)
        """L23 port errors.

        :type: P_ERRORS
        """

        self.interframe_gap = P_INTERFRAMEGAP(conn, module_id, port_id)
        """L23 port interframe gap.

        :type: P_INTERFRAMEGAP
        """

        self.max_header_length = P_MAXHEADERLENGTH(conn, module_id, port_id)
        """L23 port maximum header length.

        :type: P_MAXHEADERLENGTH
        """

        self.tpld_mode = P_TPLDMODE(conn, module_id, port_id)
        """L23 port test payload mode.

        :type: P_TPLDMODE
        """

        self.pfc_enable = P_PFCENABLE(conn, module_id, port_id)
        """L23 port Ethernet Priority Flow Control (PFC).

        :type: P_PFCENABLE
        """

        self.random_seed = P_RANDOMSEED(conn, module_id, port_id)
        """L23 port seed value.

        :type: P_RANDOMSEED
        """

        self.payload_mode = P_PAYLOADMODE(conn, module_id, port_id)
        """L23 port payload mode.

        :type: P_PAYLOADMODE
        """

        self.autotrain = P_AUTOTRAIN(conn, module_id, port_id)
        """L23 port interval between auto training packets.

        :type: P_AUTOTRAIN
        """

        self.gap_monitor = P_GAPMONITOR(conn, module_id, port_id)
        """L23 port gap monitor.

        :type: P_GAPMONITOR
        """

        self.checksum = P_CHECKSUM(conn, module_id, port_id)
        """L23 port extra payload integrity checksum.

        :type: P_CHECKSUM
        """

        self.arp_rx_table = P_ARPRXTABLE(conn, module_id, port_id)
        """L23 port ARP table.

        :type: P_ARPRXTABLE
        """

        self.ndp_rx_table = P_NDPRXTABLE(conn, module_id, port_id)
        """L23 port NDP table.

        :type: P_NDPRXTABLE
        """

        self.capturer = PortCapture(conn, module_id, port_id)
        """L23 port capturer configuration.

        :type: PortCapture
        """

        self.speed = Speed(conn, module_id, port_id)
        """L23 port speed configuration.

        :type: Speed
        """

        self.traffic = Traffic(conn, module_id, port_id)
        """L23 port traffic configuration.

        :type: Traffic
        """

        self.mix = Mix(conn, module_id, port_id)
        """L23 port IMIX configuration.

        :type: Mix
        """

        self.latency_config = LatencyConfiguration(conn, module_id, port_id)
        """L23 port latency configuration.

        self.latency_config = LatencyConfiguration(conn, module_id, port_id)
        :type:
        """

        self.rate = Rate(conn, module_id, port_id)
        """L23 port rate.

        :type: Rate
        """

        self.tx_config = TxConfiguration(conn, module_id, port_id)
        """L23 port TX configuration.

        :type: TxConfiguration
        """

        self.tx_single_pkt = TxSinglePacket(conn, module_id, port_id)
        """L23 port single-packet TX configuration.

        :type: TxSinglePacket
        """

        self.multicast = Multicast(conn, module_id, port_id)
        """L23 port multicast configuration.

        :type: Multicast
        """

        self.net_config = NetworkConfiguration(conn, module_id, port_id)
        """L23 port network configuration.

        :type: NetworkConfiguration
        """

        self.local_states = ports_state.PortL23LocalState()
        """L23 port local states.

        :type: PortL23LocalState
        """

        self.length_terms: LengthTermIndices = idx_mgr.IndexManager(
            conn,
            LengthTermIdx,
            module_id,
            port_id
        )
        """L23 port's length term index manager.

        :type: LengthTermIndices
        """

        self.match_terms: MatchTermIndices = idx_mgr.IndexManager(
            conn,
            MatchTermIdx,
            module_id,
            port_id
        )
        """L23 port's match term index manager.

        :type: MatchTermIndices
        """

    on_speed_change = functools.partialmethod(utils.on_event, P_SPEED)
    """Register a callback to the event that the port's speed changes."""

    on_traffic_change = functools.partialmethod(utils.on_event, P_TRAFFIC)
    """Register a callback to the event that the port's traffic status changes."""



from __future__ import annotations
import asyncio
import typing as t
from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.transporter.protocol.struct_request import Request
from xoa_driver.internals.core.transporter._typings import ICommand


class Green:
    @staticmethod
    def is_connected(inst: itf.IConnection) -> bool:
        return inst.is_connected

    @staticmethod
    def send(inst: itf.IConnection, data: bytes | bytearray | memoryview) -> None:
        return inst.send(data)

    @staticmethod
    def close(inst: itf.IConnection) -> None:
        return inst.close()

    @staticmethod
    async def prepare_data(inst: itf.IConnection, request: "Request") -> tuple[bytes, asyncio.Future]:
        return await inst.prepare_data(request)

    @staticmethod
    def subscribe(inst: itf.IConnection, xmc_cls: "ICommand", callback: t.Callable) -> None:
        return inst.subscribe(xmc_cls, callback)

    @staticmethod
    def on_disconnected(inst: itf.IConnection, callback: t.Callable) -> None:
        return inst.on_disconnected(callback)


class Red:
    @staticmethod
    def is_connected(inst: itf.IConnection) -> bool:
        return False

    @staticmethod
    def send(inst: itf.IConnection, data: bytes | bytearray | memoryview) -> None:
        raise ConnectionRefusedError("The instance is not valid anymore, please obtain() a new one.")

    @staticmethod
    def close(inst) -> None:
        raise ConnectionRefusedError("The instance is not valid anymore, please obtain() a new one.")

    @staticmethod
    async def prepare_data(inst, request: "Request") -> tuple[bytes, asyncio.Future]:
        raise ConnectionRefusedError("The instance is not valid anymore, please obtain() a new one.")

    @staticmethod
    def subscribe(inst, xmc_cls: "ICommand", callback: t.Callable) -> None:
        raise ConnectionRefusedError("The instance is not valid anymore, please obtain() a new one.")

    @staticmethod
    def on_disconnected(inst, callback: t.Callable) -> None:
        raise ConnectionRefusedError("The instance is not valid anymore, please obtain() a new one.")


class ConnectionTrafficLight:
    __slots__ = ("_conn", "state")

    def __init__(self, connection: itf.IConnection) -> None:
        self._conn = connection
        self.state = Green

    def set_outdated(self) -> None:
        self.state = Red

    @property
    def is_connected(self) -> bool:
        return self.state.is_connected(self._conn)

    def send(self, data: bytes | bytearray | memoryview) -> None:
        return self.state.send(self._conn, data)

    def close(self) -> None:
        return self.state.close(self._conn)

    async def prepare_data(self, request: "Request") -> tuple[bytes, asyncio.Future]:
        return await self.state.prepare_data(self._conn, request)

    def subscribe(self, xmc_cls: "ICommand", callback: t.Callable) -> None:
        return self.state.subscribe(self._conn, xmc_cls, callback)

    def on_disconnected(self, callback: t.Callable) -> None:
        return self.state.on_disconnected(self._conn, callback)



import asyncio
import functools
from typing import TYPE_CHECKING

from xoa_driver.internals.commands import (
    P_CAPABILITIES,
    P_INTERFACE,
    P_STATUS,
    P_TXENABLE,
    P_LOADMODE,
    P_EMULATE,
    P_CAPABILITIES_EXT,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.hli_v1.ports import base_port
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.state_storage import ports_state
from ..pcs_pma_ijkl_chimera import PcsPma
from .pe_custom_distribution import CustomDistributions
from .port_emulation import ChimeraPE


class PortChimera(base_port.BasePort[ports_state.PortChimeraLocalState]):
    """Chimera port for traffic impairment."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.capabilities = P_CAPABILITIES(conn, module_id, port_id)
        """Chimera port capabilities.

        :type: P_CAPABILITIES
        """

        self.capabilities_ext = P_CAPABILITIES_EXT(conn, module_id, port_id)
        """Chimera port capabilities ext.

        :type: P_CAPABILITIES_EXT
        """

        self.interface = P_INTERFACE(conn, module_id, port_id)
        """Physical interface type of the Chimera port.

        :type: P_INTERFACE
        """

        self.status = P_STATUS(conn, module_id, port_id)
        """Chimera port's received optical signal level'.

        :type: P_STATUS
        """

        self.tx_enable = P_TXENABLE(conn, module_id, port_id)
        """Enabling Chimera port TX.

        :type: P_TXENABLE
        """

        self.load_mode = P_LOADMODE(conn, module_id, port_id)
        """Load mode of the Chimera port.

        :type: P_LOADMODE
        """

        self.emulation = ChimeraPE(self._conn, *self.kind)
        """Chimera Port Emulation """

        self.emulate = P_EMULATE(conn, module_id, port_id)
        """Chimera port emulation control.

        :type: P_EMULATE
        """

        self.custom_distributions = CustomDistributions(conn, module_id, port_id)
        """Custom distributions."""

        self.pcs_pma = PcsPma(conn, self)
        """"PCS/PMA settings."""

        self._local_states = ports_state.PortChimeraLocalState()
        """Local states of the Chimera port."""

    @property
    def info(self) -> ports_state.PortChimeraLocalState:
        return self._local_states

    async def _setup(self):
        await asyncio.gather(
            self._local_states.initiate(self),
            self.custom_distributions.server_sync(),
            self.emulation
        )
        self._local_states.register_subscriptions(self)
        return self

    on_interface_change = functools.partialmethod(utils.on_event, P_INTERFACE)
    """Register a callback to the event that the port's interface type changes."""

    on_emulate_change = functools.partialmethod(utils.on_event, P_EMULATE)
    """Register a callback to the event that the port's emulation state changes."""


class PChi100G5S2P(PortChimera):
    """Impairment port on Chi-100G-5S-2P module.
    """
    ...


class PChi100G5S2P_b(PortChimera):
    """Impairment port on Chi-100G-5S-2P[b] module.
    """
    ...


class PChi40G2S2P(PortChimera):
    """Impairment port on Chi-40G-2S-2P module.
    """
    ...



from __future__ import annotations
import typing as t

if t.TYPE_CHECKING:
    from .__logger import TransportationLogger


class StateOff:

    @staticmethod
    def setup(connid: str, logger) -> None:
        return None

    @staticmethod
    def info(inst: "TransportationLogger", msg: t.Any) -> None:
        return None

    @staticmethod
    def error(inst: "TransportationLogger", msg: t.Any) -> None:
        return None

    @staticmethod
    def debug_request(inst: "TransportationLogger", request: object) -> None:
        return None

    @staticmethod
    def debug_response(inst: "TransportationLogger", response: object) -> None:
        return None

    @staticmethod
    def debug_push(inst: "TransportationLogger", response: object) -> None:
        return None



from typing import (
    TYPE_CHECKING,
    Dict,
    List,
)
from collections import UserDict
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    PEC_INDICES,
    PEC_VAL,
    PEC_COMMENT,
    PEC_DELETE,
    PEC_DISTTYPE,
)

from xoa_driver.internals.utils.indices import observer
from xoa_driver.enums import OnOff


class CustomDistribution:
    """Custom distribution"""

    def __init__(self, observer: "observer.IndicesObserver", conn: "itf.IConnection", module_id: int, port_id: int, custom_distribution_index: int) -> None:
        self.__observer = observer
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.custom_distribution_index = custom_distribution_index
        self.definition = PEC_VAL(conn, module_id, port_id, custom_distribution_index)
        """Custom distribution definition.
        Representation of PEC_VAL
        """

        self.comment = PEC_COMMENT(conn, module_id, port_id, custom_distribution_index)
        """Custom distribution description.
        Representation of PEC_COMMENT
        """

        self.type = PEC_DISTTYPE(conn, module_id, port_id, custom_distribution_index)
        """Custom distribution type.
        Representation of PEC_DISTTYPE
        """

    async def delete(self) -> None:
        """
        Deleting an existing Custom Distribution
        """

        await PEC_DELETE(
            self.__conn,
            self.__module_id,
            self.__port_id,
            self.custom_distribution_index
        ).set()
        self.__observer.notify(observer.IndexEvents.DEL, self)


class CustomDistributions(UserDict):
    """Custom distributions"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.data: Dict[int, CustomDistribution] = {}
        self.__observer = observer.IndicesObserver()
        self.__observer.subscribe(
            observer.IndexEvents.DEL,
            self.__remove_from_slot
        )

    async def server_sync(self) -> None:
        """Sync the indices with xenaserver"""
        _resp = await PEC_INDICES(self.__conn, self.__module_id, self.__port_id).get()
        self.data = {
            idx: CustomDistribution(
                self.__observer,
                self.__conn,
                self.__module_id,
                self.__port_id,
                idx
            )
            for idx in _resp.indexations
        }

    def __remove_from_slot(self, index_inst: "CustomDistribution") -> None:
        # throws ValueError if element is not exists in list of indices
        del self.data[index_inst.custom_distribution_index]

    def __setitem__(self, key, value):
        raise NotImplementedError("Only support assign item by 'add' method")

    async def remove(self, custom_distribution_index: int) -> None:
        """Remove a index from port"""
        await self.data[custom_distribution_index].delete()

    async def __get_available_custom_distribution_index(self) -> int:
        await self.server_sync()
        if len(self.keys()) == 40:
            raise ValueError("The server was full of custom distributions.")
        return next(i for i in range(1, 41) if i not in self.keys())

    async def add(self, linear: OnOff, entry_count: int, data_x: List[int], comment: str) -> "CustomDistribution":
        cdi = await self.__get_available_custom_distribution_index()
        cd = CustomDistribution(
            self.__observer,
            self.__conn,
            self.__module_id,
            self.__port_id,
            custom_distribution_index=cdi,
        )
        await cd.definition.set(linear=linear, symmetric=OnOff.OFF, entry_count=entry_count, data_x=data_x)
        await cd.comment.set(comment)
        self.data[cdi] = cd
        return cd



import functools
from typing import TYPE_CHECKING
from typing_extensions import Self
from xoa_driver.internals.commands import (
    # P_FAULTSIGNALING,
    P_DYNAMIC,
)
from xoa_driver.internals.utils import attributes as utils
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from .bases.port_l23_genuine import BasePortL23Genuine
from .fault_jkl import Fault
from .pcs_pma_ijkl_chimera import PcsPma as PcsPma1
from .pcs_pma_ghijkl import (
    PcsPma as PcsPma2,
    SerDes,
)
from .pcs_pma_l import PcsPma as PcsPma3


class PcsPma(PcsPma1, PcsPma2, PcsPma3):
    def __init__(self, conn: "itf.IConnection", port) -> None:
        PcsPma1.__init__(self, conn, port)
        PcsPma2.__init__(self, conn, port)
        PcsPma3.__init__(self, conn, port)


class FamilyL(BasePortL23Genuine):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.dynamic = P_DYNAMIC(conn, module_id, port_id)
        """L23 port's dynamic traffic change.
        Representation of P_DYNAMIC
        """

        self.fault = Fault(conn, module_id, port_id)

    async def _setup(self) -> Self:
        await super()._setup()
        self.pcs_pma = PcsPma(self._conn, self)
        self.serdes = tuple(
            SerDes(self._conn, *self.kind, serdes_xindex=serdes_xindex)
            for serdes_xindex in range(self.info.capabilities.serdes_count)
        )
        return self

    on_dynamic_change = functools.partialmethod(utils.on_event, P_DYNAMIC)
    """Register a callback to the event that the port's dynamic traffic setting changes."""


class PThor400G7S1P_b(FamilyL):
    """L23 port on Thor-400G-7S-1P[b] module.
    """
    ...


class PThor400G7S1P_c(FamilyL):
    """L23 port on Thor-400G-7S-1P[c] module.
    """
    ...


class PThor400G7S1P_d(FamilyL):
    """L23 port on Thor-400G-7S-1P[d] module.
    """
    ...



from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    C_FILESTART,
    C_FILEDATA,
    C_FILEFINISH
)


class UploadFile:
    """File uploading functions of the Valkyrie tester."""

    def __init__(self, conn: "itf.IConnection") -> None:
        self.start = C_FILESTART(conn)
        """Start uploading file.
        Representation of C_FILESTART
        """

        self.data = C_FILEDATA(conn)
        """Uploading a file fragment.
        Representation of C_FILEDATA
        """

        self.finish = C_FILEFINISH(conn)
        """Finish uploading file.
        Representation of C_FILEFINISH
        """



from __future__ import annotations
import asyncio
import typing as t
from xoa_driver import enums, testers
from xoa_driver.utils import apply
from xoa_driver.internals.hli_v2.ports.port_l23.family_l import FamilyL
from xoa_driver.internals.hli_v2.ports.port_l23.family_l1 import FamilyL1
from xoa_driver.ports import GenericAnyPort
from xoa_driver.modules import GenericAnyModule, GenericL23Module, ModuleChimera
from xoa_driver.testers import GenericAnyTester
from .exceptions import (
    NotSupportMedia,
    NotSupportPortSpeed,
)
from .tools import MODULE_EOL_INFO
from itertools import chain  # type: ignore[Pylance false warning]
from datetime import datetime

PcsPmaSupported = (FamilyL, FamilyL1)
AutoNegSupported = (FamilyL, FamilyL1)
LinkTrainingSupported = FamilyL


# region Testers
async def reserve_tester(tester: GenericAnyTester, force: bool = True) -> None:
    """
    .. versionadded:: 1.1

    Reserve a tester regardless whether it is owned by others or not.

    :param tester: The tester to reserve
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :param force: Should force reserve the tester
    :type force: boolean
    :return:
    :rtype: None
    """
    r = await tester.reservation.get()
    if force and r.operation == enums.ReservedStatus.RESERVED_BY_OTHER:
        await tester.reservation.set_relinquish()
        await asyncio.gather(*(free_module(m, True) for m in tester.modules))
        await tester.reservation.set_reserve()
    elif r.operation == enums.ReservedStatus.RELEASED:
        await tester.reservation.set_reserve()


async def free_tester(
        tester: GenericAnyTester, 
        should_free_modules_ports: bool = False,
        ) -> None:
    """
    .. versionadded:: 1.1

    Free a tester. If the tester is reserved by you, release the tester. If the tester is reserved by others, relinquish the tester. The tester should have no owner afterwards.

    :param tester: The tester to free
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :param should_free_modules_ports: should modules and ports also be freed, defaults to False
    :type should_free_modules_ports: bool, optional
    :return:
    :rtype: None
    """
    r = await tester.reservation.get()
    if r.operation == enums.ReservedStatus.RESERVED_BY_OTHER:
        await tester.reservation.set_relinquish()
    elif r.operation == enums.ReservedStatus.RESERVED_BY_YOU:
        await tester.reservation.set_release()
    if should_free_modules_ports:
        await asyncio.gather(*(free_module(m, True) for m in tester.modules))


# endregion


# region Modules


def get_module(tester: GenericAnyTester, module_id: int) -> GenericAnyModule:
    """
    .. versionadded:: 1.1

    Get a module object of the tester.

    :param tester: The tester object
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :param module_id: the index id of the module
    :type module_id: int
    :raises NoSuchModuleError: No such a module index on the tester
    :return: module object
    :rtype: :class:`~xoa_driver.modules.GenericAnyModule`
    """
    return tester.modules.obtain(module_id)


def get_modules(tester: GenericAnyTester) -> tuple[GenericAnyModule, ...]:
    """
    .. versionadded:: 1.1

    Get all modules of the tester

    :param tester: The tester object
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :return: List of module objects
    :rtype: tuple[GenericAnyModule]
    """
    return tuple(tester.modules)


async def reserve_module(module: GenericAnyModule, force: bool = True) -> None:
    """
    .. versionadded:: 1.1

    Reserve a module regardless whether it is owned by others or not.

    :param module: The module to reserve
    :type module: :class:`~xoa_driver.modules.GenericAnyModule`
    :param force: Should force reserve the module, defaults to True
    :type force: boolean
    :return:
    :rtype: None
    """
    r = await module.reservation.get()
    if force and r.operation == enums.ReservedStatus.RESERVED_BY_OTHER:
        await free_module(module, True)
        await module.reservation.set_reserve()
    elif r.operation == enums.ReservedStatus.RELEASED:
        await module.reservation.set_reserve()


async def free_module(
    module: GenericAnyModule, should_free_ports: bool = False
) -> None:
    """
    .. versionadded:: 1.2

    Free a module. If the module is reserved by you, release the module. If the module is reserved by others, relinquish the module. The module should have no owner afterwards.

    :param module: The module to free
    :type module: :class:`~xoa_driver.modules.GenericAnyModule`
    :param should_free_ports: should ports also be freed, defaults to False
    :type should_free_ports: bool, optional
    :return:
    :rtype: None
    """
    r = await module.reservation.get()
    if r.operation == enums.ReservedStatus.RESERVED_BY_OTHER:
        await module.reservation.set_relinquish()
    elif r.operation == enums.ReservedStatus.RESERVED_BY_YOU:
        await module.reservation.set_release()
    if should_free_ports:
        await free_ports(*module.ports)


def get_module_supported_media(
    module: GenericL23Module | ModuleChimera,
) -> list[dict[str, t.Any]]:
    """
    .. versionadded:: 1.3

    Get a list of supported media, port speed and count of the module.

    :param module: The module object
    :type module: GenericAnyModule
    :return: List of supported media, port speed and count
    :rtype: list[dict[str, t.Any]]
    """
    supported_media_list = []
    item = {}

    for media_item in module.info.media_info_list:  # type: ignore
        for sub_item in media_item.available_speeds:
            item = dict()
            item["media"] = media_item.cage_type
            item["port_count"] = sub_item.port_count
            item["port_speed"] = sub_item.port_speed
            supported_media_list.append(item)

    return supported_media_list


async def set_module_media_config(
    module: t.Union[GenericL23Module, ModuleChimera],
    media: enums.MediaConfigurationType,
    force: bool = True,
) -> None:
    """
    .. versionadded:: 1.3

    Set module's media configuration.

    :param module: The module object
    :type module: GenericAnyModule
    :param media: Target media for the module
    :type media: enums.MediaConfigurationType
    :param force: Should reserve the module by force, defaults to True
    :type force: bool, optional
    :raises NotSupportMedia: The module does not support this media type
    :return:
    :rtype:
    """

    # reserve the module first
    await reserve_module(module, force)

    # get the supported media
    supported_media_list = get_module_supported_media(module)

    # set the module media if the target media is found in supported media
    for item in supported_media_list:
        if item["media"] == media:
            await module.media.set(media_config=media)
            return None

    # raise exception is the target media is not found in the supported media
    raise NotSupportMedia(module)


async def set_module_port_config(
    module: t.Union[GenericL23Module, ModuleChimera],
    port_count: int,
    port_speed: int,
    force: bool = True,
) -> None:
    """
    .. versionadded:: 1.3

    Set module's port-speed configuration

    :param module: The module object
    :type module: t.Union[GenericL23Module, ModuleChimera]
    :param port_count: The port count
    :type port_count: int
    :param port_speed: The port speed in Mbps, e.g. 40000 for 40G
    :type port_speed: int
    :param force: Should reserve the module by force, defaults to True
    :type force: bool, optional
    :raises NotSupportPortSpeed: The module does not support the port-speed configuration under its current media configuration
    :return:
    :rtype:
    """

    # reserve the module first
    await free_module(module, True)
    await reserve_module(module, force)

    # get the supported media by the module
    supported_media_list = get_module_supported_media(module)

    # get the current media of the module
    reply = await module.media.get()
    current_media = reply.media_config

    # set the module port speed if we can find the port-speed in the corresponding media
    for item in supported_media_list:
        if all(
            (
                item["media"] == enums.MediaConfigurationType(current_media),
                item["port_count"] == port_count,
                item["port_speed"] == port_speed,
            )
        ):
            portspeed_list = [port_count] + port_count * [port_speed]
            await module.cfp.config.set(portspeed_list=portspeed_list)
            await free_module(module, False)
            return None
    raise NotSupportPortSpeed(module)


async def get_module_eol_date(module: GenericAnyModule) -> str:
    """
    .. versionadded:: 1.3

    Get module's End-of-Life date

    :param module: The module object
    :type module: GenericAnyModule
    :return: Module's EOL date
    :rtype: str
    """
    resp = await module.serial_number.get()
    module_key = str(resp.serial_number)[-2:]
    return MODULE_EOL_INFO.get(module_key, "2999-01-01")


async def get_module_eol_days(module: GenericAnyModule) -> int:
    """
    .. versionadded:: 1.3

    Get days until module's End-of-Life date

    :param module: The module object
    :type module: GenericAnyModule
    :return: days until module's End-of-Life date
    :rtype: int
    """
    eol_string = await get_module_eol_date(module)
    date1 = datetime.now()
    date2 = datetime.strptime(eol_string, "%Y-%M-%d")
    timedelta = date2 - date1
    return timedelta.days


# endregion


# region Ports


def get_all_ports(tester: GenericAnyTester) -> tuple[GenericAnyPort, ...]:
    """
    .. versionadded:: 1.1

    Get all ports of the tester

    :param tester: The tester object
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :return: List of port objects
    :rtype: tuple[GenericAnyPort]
    """
    all_ports_ = (m.ports for m in get_modules(tester))
    return tuple(chain.from_iterable(all_ports_))


def get_ports(tester: GenericAnyTester, module_id: int) -> tuple[GenericAnyPort, ...]:
    """
    .. versionadded:: 1.1

    Get all ports of the module

    :param tester: The tester object
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :param module_id: The module index
    :type module_id: int
    :return: List of port objects
    :rtype: tuple[GenericAnyPort]
    """
    module = get_module(tester, module_id)
    return tuple(module.ports)


def get_port(tester: GenericAnyTester, module_id: int, port_id: int) -> GenericAnyPort:
    """
    .. versionadded:: 1.1

    Get a port of the module

    :param tester: The tester object
    :type tester: :class:`~xoa_driver.testers.GenericAnyTester`
    :param module_id: The module index
    :type module_id: int
    :param port_id: The port index
    :type port_id: int
    :raises NoSuchPortError: No port found with the index
    :return: The port object
    :rtype: GenericAnyPort
    """
    module = get_module(tester, module_id)
    return module.ports.obtain(port_id)


async def reserve_port(port: GenericAnyPort, force: bool = True) -> None:
    """
    .. versionadded:: 1.1

    Reserve a port regardless whether it is owned by others or not.

    :param port: The port to reserve
    :type port: GenericAnyPort
    :param force: Should force reserve the port
    :type force: boolean
    :return:
    :rtype: None
    """
    r = await port.reservation.get()
    if force and r.status == enums.ReservedStatus.RESERVED_BY_OTHER:
        await apply(
            port.reservation.set_relinquish(),
            port.reservation.set_reserve(),
        )
    elif r.status == enums.ReservedStatus.RELEASED:
        await port.reservation.set_reserve()


async def reset_port(port: GenericAnyPort) -> None:
    """
    .. versionadded:: 1.1

    Reserve and reset a port

    :param port: The port to reset
    :type port: GenericAnyPort
    :return:
    :rtype: None
    """
    await reserve_port(port, False)
    await port.reset.set()


async def free_port(port: GenericAnyPort) -> None:
    """
    .. versionadded:: 1.1

    Free a port. If the port is reserved by you, release the port. If the port is reserved by others, relinquish the port. The port should have no owner afterwards.

    :param port: The port to free
    :type port: GenericAnyPort
    :return:
    :rtype: None
    """
    r = await port.reservation.get()
    if r.status == enums.ReservedStatus.RESERVED_BY_OTHER:
        await port.reservation.set_relinquish()
    elif r.status == enums.ReservedStatus.RESERVED_BY_YOU:
        await port.reservation.set_release()


async def free_ports(*ports: GenericAnyPort) -> None:
    """
    .. versionadded:: 1.1

    Free all ports on a module.

    :param port: The port to free
    :type port: GenericAnyPort
    """
    await asyncio.gather(*(free_port(port=p) for p in ports))



# endregion


# region Streams
async def remove_streams(port: GenericAnyPort) -> None:
    """
    .. versionadded:: 2.1

    Remove all streams on a port.

    :param module: The port object
    :type module: GenericAnyPort
    """
    await port.streams.server_sync()
    await asyncio.gather(*(s.delete() for s in port.streams))


# endregion

__all__ = (
    "free_module",
    "free_port",
    "free_ports",
    "free_tester",
    "get_all_ports",
    "get_module",
    "get_module_eol_date",
    "get_module_eol_days",
    "get_module_supported_media",
    "get_modules",
    "get_port",
    "get_ports",
    "reserve_module",
    "reserve_port",
    "reserve_tester",
    "reset_port",
    "set_module_media_config",
    "set_module_port_config",
    "remove_streams"
)



from typing import (
    List,
    Type,
    TypeVar,
    TYPE_CHECKING,
)
from xoa_driver.internals.commands import (
    PM_INDICES,
    PM_CREATE,
    PM_DELETE,
    PM_PROTOCOL,
    PM_POSITION,
    PM_MATCH,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
from xoa_driver.internals.utils.indices import observer as idx_obs
from .base_index import BaseIndex

MT = TypeVar("MT")


class MatchTermIdx(BaseIndex):
    """L23 Match Term Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)

        self.protocol = PM_PROTOCOL(conn, *kind)
        """The protocol segments assumed on the packets received on the port.

        :type: PM_PROTOCOL
        """

        self.position = PM_POSITION(conn, *kind)
        """The position within each received packet where content matching begins for the port.

        :type: PM_POSITION
        """

        self.match = PM_MATCH(conn, *kind)
        """The value that must be found at the match term position for packets received on the port.

        :type: PM_MATCH
        """

    async def delete(self):
        await PM_DELETE(self._conn, *self.kind).set()
        """Delete the match term.
        
        :type: PM_DELETE
        """

        self._observer.notify(idx_obs.IndexEvents.DEL, self)

    @classmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        resp = await PM_INDICES(conn, module_id, port_id).get()
        return list(resp.match_term_xindices)

    @classmethod
    async def _new(cls: Type[MT], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> MT:
        await PM_CREATE(conn, *kind).set()
        return cls(conn, kind, observer)



from __future__ import annotations
import asyncio
import functools
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Iterable,
    NamedTuple,
    Type,
    TypeVar,
    Callable,
    Union,
    cast,
)
from collections import OrderedDict

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    M_REVISION,
    M_MODEL
)
from .abc import AbcResourcesManager
from .exceptions import (
    NoSuchModuleError,
    WrongTesterError
)

# region Module Init dataset


class ModuleData(NamedTuple):
    """Base module data."""
    module_id: int
    ports_count: int
    revision: str


class L23ModuleData(ModuleData):
    """L23 module data"""
    @classmethod
    async def create(cls, conn, mid: int, p_cnt: int) -> "L23ModuleData":
        """Return the module name of the L23 module."""
        r = await M_REVISION(conn, mid).get()
        return cls(
            module_id=mid,
            ports_count=p_cnt,
            revision=r.revision
        )


class L47ModuleData(ModuleData):
    """L47 module data"""
    @classmethod
    async def create(cls, conn, mid: int, p_cnt: int) -> "L47ModuleData":
        """Return the module name of the L47 module."""
        r = await M_MODEL(conn, mid).get()
        return cls(
            module_id=mid,
            ports_count=p_cnt,
            revision=r.model
        )

# endregion


MT = TypeVar("MT")
MTypeObtainer = Callable[[str], Type]
ModuleDataType = Union[Type["L23ModuleData"], Type["L47ModuleData"]]


class ModulesManager(AbcResourcesManager[MT]):
    """Modules Manager"""

    __slots__ = ("_conn", "__m_types_obtainer")

    def __init__(self, conn: "itf.IConnection", m_types_obtainer: MTypeObtainer) -> None:
        super().__init__()
        self._conn = conn
        self.__m_types_obtainer = m_types_obtainer

    def obtain(self, key: int) -> MT:
        try:
            return super().obtain(key)
        except KeyError:
            raise NoSuchModuleError(key)

    def obtain_multiple(self, *keys: int) -> tuple[MT, ...]:
        try:
            return super().obtain_multiple(*keys)
        except KeyError as err:
            raise NoSuchModuleError(err.args[0])

    async def fill(self, ports_count: list[int], module_type: ModuleDataType) -> None:
        """Method for create and fill in."""
        assert not self._lock, "Method <fill> can be called only once."
        module_builder = (
            module_type.create(self._conn, slot_id, p_count)
            for slot_id, p_count in enumerate(ports_count)
            if p_count > 0
        )
        identities: Iterable[ModuleData] = await asyncio.gather(*module_builder)

        self._items = OrderedDict(
            (
                idnt.module_id,
                self.__m_types_obtainer(idnt.revision)(self._conn, idnt)
            )
            for idnt in identities
        )
        if len(self) == 0:
            raise WrongTesterError()
        coros = cast(Iterable[Awaitable], self._items.values())
        await asyncio.gather(*coros)

    fill_l23 = functools.partialmethod(fill, module_type=L23ModuleData)
    fill_l47 = functools.partialmethod(fill, module_type=L47ModuleData)



from __future__ import annotations
from dataclasses import dataclass
import typing

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpHex,
    XmpSequence,
    XmpInt,
    Hex,
)


@register_command
@dataclass
class PX_RW:
    """
    Provides read and write access to the register interface supported by the port transceiver. It is possible to both read and write register values.
    """

    code: typing.ClassVar[int] = 501
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _page_xindex: int
    _register_xaddress: int

    class GetDataAttr(ResponseBodyStruct):
        value: Hex = field(XmpHex(size=4))
        """4 hex bytes, register value of the port transceiver"""

    class SetDataAttr(RequestBodyStruct):
        value: Hex = field(XmpHex(size=4))
        """4 hex bytes, register value of the port transceiver"""

    def get(self) -> Token[GetDataAttr]:
        """Get the register value of a transceiver.

        :return: the register value of a transceiver
        :rtype: PX_RW.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress]))

    def set(self, value: Hex) -> Token[None]:
        """Set the register value of a transceiver.

        :param value: register value of a transceiver
        :type value: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress], value=value))


@register_command
@dataclass
class PX_RW_SEQ:
    """
    :term:`I2C<I2C>` sequential access to a transceiver's register.
    When invoked, the ``<byte_count>`` number of bytes will be read or written in one I2C transaction,
    in which the ``<value>`` is read or written with only a single register address setup.
    A subsequent invocation will perform a second I2C transaction in the same manner.
    ``<_page_xindex>``: the transceiver page address, integer, 0-255.
    ``<_register_xaddress>``: the address within the page, integer, 0-255.
    """

    code: typing.ClassVar[int] = 503
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _page_xindex: int
    _register_xaddress: int
    _byte_xcount: int

    class GetDataAttr(ResponseBodyStruct):
        value: Hex = field(XmpHex())
        """the bytes to be read or written in one I2C transaction. The number of bytes in the ``<value>`` equals ``<byte_count>``."""

    class SetDataAttr(RequestBodyStruct):
        value: Hex = field(XmpHex())
        """the bytes to be read or written in one I2C transaction. The number of bytes in the ``<value>`` equals ``<byte_count>``."""

    def get(self) -> Token[GetDataAttr]:
        """Get the register value of a transceiver in one I2C transaction.

        :return: the register value of a transceiver
        :rtype: PX_RW_SEQ.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress, self._byte_xcount]))

    def set(self, value: Hex) -> Token[None]:
        """Set the register value of a transceiver in one I2C transaction.

        :param value: register value of a transceiver
        :type value: Hex
        """

        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._page_xindex, self._register_xaddress, self._byte_xcount],
                value=value
            )
        )


@register_command
@dataclass
class PX_MII:
    """Provides access to the register interface supported by the media-independent interface (MII) transceiver. It
    is possible to both read and write register values."""

    code: typing.ClassVar[int] = 537
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _register_xaddress: int

    class GetDataAttr(ResponseBodyStruct):
        value: Hex = field(XmpHex(size=2))
        """2 hex bytes, register value of the transceiver"""

    class SetDataAttr(RequestBodyStruct):
        value: Hex = field(XmpHex(size=2))
        """2 hex bytes, register value of the transceiver"""

    def get(self) -> Token[GetDataAttr]:
        """Get the register value of a transceiver.

        :return: the register value of a transceiver
        :rtype: PX_MII.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._register_xaddress]))

    def set(self, value: Hex) -> Token[None]:
        """Set the register value of a transceiver.

        :param value: register value of a transceiver
        :type value: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._register_xaddress], value=value))


@register_command
@dataclass
class PX_TEMPERATURE:
    """
    Transceiver temperature in degrees Celsius.
    """

    code: typing.ClassVar[int] = 538
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        integral_part: int = field(XmpByte())
        """byte, temperature value before the decimal digit."""
        fractional_part: int = field(XmpByte())
        """byte, 1/256th of a degree Celsius after the decimal digit."""

    def get(self) -> Token[GetDataAttr]:
        """Get transceiver temperature in degrees Celsius.

        :return: temperature value before the decimal digit, and 1/256th of a degree Celsius after the decimal digit.
        :rtype: PX_TEMPERATURE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

@register_command
@dataclass
class PX_I2C_CONFIG:
    """
    Set the get the access speed on a transceiver I2C access in the unit of KHz. Default to 100. When the transceiver is plugged out and in again, the speed will be reset to the default value 100. The speed has a minimum and a maximum, which can be obtained from P_CAPABILITIES. The I2C speed configuration will not be included in the port configuration file (.xpc). When you load a port configuration to a port, the transceiver I2C access speed will be reset to default 100.
    """

    code: typing.ClassVar[int] = 539
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        frequency: int = field(XmpInt())
        """integer, frequency in kHz, default is 100."""
    
    class SetDataAttr(RequestBodyStruct):
        frequency: int = field(XmpInt())
        """integer, frequency in kHz, default is 100."""

    def get(self) -> Token[GetDataAttr]:
        """Get the speed on a transceiver I2C access in the unit of KHz.

        :return: frequency in kHz.
        :rtype: PX_I2C_CONFIG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))
    
    def set(self, frequency: int) -> Token[None]:
        """Set the speed on a transceiver I2C access in the unit of KHz.

        :param frequency: frequency in kHz
        :type frequency: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, frequency=frequency))


from typing import TYPE_CHECKING
from .bases.port_l23 import BasePortL23
from .bases.port_reception_statistics import PortReceptionStatistics
from .bases.port_transmission_statistics import PortTransmissionStatistics

from xoa_driver.internals.commands import (
    P_MDIXMODE,
    # P_ENGINENAMES, # TODO: need to implement
    # P_ENGINELOAD, # TODO: need to implement
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.utils.indices import index_manager as idx_mgr
from xoa_driver.internals.hli_v2.indices.streams.base_stream import BaseStreamIdx
from xoa_driver.internals.hli_v2.indices.filter.base_filter import BaseFilterIdx
from xoa_driver.internals.state_storage import ports_state

VEStreamIndices = idx_mgr.IndexManager[BaseStreamIdx]
VEFilterIndices = idx_mgr.IndexManager[BaseFilterIdx]


class Engine:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        # self.names = P_ENGINENAMES(conn, module_id, port_id)
        # self.load = P_ENGINELOAD(conn, module_id, port_id)
        ...


class L23VEPortStatistics:
    """L23 VE port statistics"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = PortReceptionStatistics(conn, module_id, port_id)
        """L23 VE port's RX statistics."""

        self.tx = PortTransmissionStatistics(conn, module_id, port_id)
        """L23 VE port's TX statistics."""


class PortL23VE(BasePortL23):
    """L23 VE port"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)

        self._local_states = ports_state.PortL23LocalState()

        self.mdix_mode = P_MDIXMODE(conn, module_id, port_id)
        """MDI/MDIX mode.
        Representation of P_MDIXMODE
        """

        self.engine = Engine(conn, module_id, port_id)
        """Engine is not supported yet."""

        self.statistics = L23VEPortStatistics(conn, module_id, port_id)

        self.streams: VEStreamIndices = idx_mgr.IndexManager(
            conn,
            BaseStreamIdx,
            module_id,
            port_id
        )
        """L23 VE port's stream index manager."""

        self.filters: VEFilterIndices = idx_mgr.IndexManager(
            conn,
            BaseFilterIdx,
            module_id,
            port_id
        )
        """L23 VE port's filter index manager."""

    @property
    def info(self) -> ports_state.PortL23LocalState:
        return self._local_states

    async def _setup(self):
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        return self



from __future__ import annotations
from typing import Generator
from . import _constants as const


def repr_bytes(data: bytes) -> list[str]:
    return data.hex(",").split(",")


def get_code_str(x) -> Generator[int | str, None, None]:
    code, ty = (
        x.header.cmd_code,
        x.header.cmd_type,
    )
    yield const.CommandType(ty).name
    if ty == const.CommandType.COMMAND_STATUS:
        yield const.CommandStatus(code).name
    else:
        yield code


def format_repr(obj) -> str:
    ty_str, code_str = get_code_str(obj)
    return (
        f"{str(obj.header.request_identifier):5s} "
        f"{str(obj.header.module_index):3s} "
        f"{str(obj.header.port_index):3s} "
        f"{str(obj.index_values):10s} "
        f"{str(obj.class_name):25s} "
        f"{str(code_str):6s} "
        f"{str(ty_str):15s} "
        f"{obj.values}"
    )


def format_str(obj) -> str:
    bin_str = repr_bytes(bytes(obj))
    ty_str, code_str = get_code_str(obj)
    obj_name = type(obj).__name__
    if obj_name == 'Response':
        cmd_p = 'Replied' if obj.header.request_identifier != 0 else 'Pushed'
    else:
        cmd_p = [code_str, ty_str]

    return "\n" + "\n".join(
        (
            f"{obj_name}              : {bin_str}",
            f"class_name           : {obj.class_name}",
            f"magic_word           : {obj.header.magic_word}",
            f"number_of_indices    : {obj.header.number_of_indices}",
            f"number_of_value_bytes: {obj.header.number_of_value_bytes}",
            f"command_parameter    : {obj.header.command_parameter}:{cmd_p}",
            f"module_index         : {obj.header.module_index}",
            f"port_index           : {obj.header.port_index}",
            f"request_identifier   : {obj.header.request_identifier}",
            f"index_values         : {obj.index_values}",
            f"values               : {obj.values}",
        )
    )



from __future__ import annotations
from .transporter.protocol.struct_request import Request
from .transporter.protocol._constants import CommandType
from .transporter._typings import (
    ICmdOnlySet,
    ICmdOnlyGet,
)


def build_set_request(cls: ICmdOnlySet, **kwargs) -> Request:
    indices = kwargs.pop("indices", [])
    module = kwargs.pop("module", None)
    port = kwargs.pop("port", None)
    req_values = cls.SetDataAttr(**kwargs)
    return Request(
        class_name=type(cls).__name__,
        cmd_type=CommandType.COMMAND_VALUE,
        cmd_code=cls.code,
        module_index=module,
        port_index=port,
        indices=indices,
        values=req_values
    )


def build_get_request(cls: ICmdOnlyGet, **kwargs) -> Request:
    indices = kwargs.pop("indices", [])
    module = kwargs.pop("module", None)
    port = kwargs.pop("port", None)
    req_values = None
    return Request(
        class_name=type(cls).__name__,
        cmd_type=CommandType.COMMAND_QUERY,
        cmd_code=cls.code,
        module_index=module,
        port_index=port,
        indices=indices,
        values=req_values
    )



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands.enums import ImpairmentTypeIndex
from xoa_driver.internals.commands import (
    PED_SCHEDULE,
    PED_ONESHOTSTATUS,
    PED_OFF,
    PED_FIXED,
    PED_RANDOM,
    PED_BER,
    PED_FIXEDBURST,
    PED_RANDOMBURST,
    PED_GE,
    PED_UNI,
    PED_GAUSS,
    PED_POISSON,
    PED_GAMMA,
    PED_CUST,
    PED_CONST,
    PED_ACCBURST,
    PED_STEP,
    PED_ENABLE,
)


class ImpairmentDistribution:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, impairment_type_index: "ImpairmentTypeIndex") -> None:
        self.off = PED_OFF(conn, module_id, port_id, flow_index, impairment_type_index)
        """Impairments Distribution to OFF.
        Representation of PED_OFF
        """

        self.fixed_rate = PED_FIXED(conn, module_id, port_id, flow_index, impairment_type_index)
        """Fixed Rate distribution configuration.
        Representation of PED_FIXED
        """

        self.random_rate = PED_RANDOM(conn, module_id, port_id, flow_index, impairment_type_index)
        """Random Rate distribution configuration.
        Representation of PED_RANDOM
        """

        self.bit_error_rate = PED_BER(conn, module_id, port_id, flow_index, impairment_type_index)
        """Bit Error Rate distribution configuration.
        Representation of PED_BER
        """

        self.random_burst = PED_RANDOMBURST(conn, module_id, port_id, flow_index, impairment_type_index)
        """Random Burst configuration.
        Representation of PED_RANDOMBURST
        """

        self.ge = PED_GE(conn, module_id, port_id, flow_index, impairment_type_index)
        """Gilbert-Elliot distribution configuration.
        Representation of PED_GE
        """

        self.uniform = PED_UNI(conn, module_id, port_id, flow_index, impairment_type_index)
        """Uniform distribution configuration.
        Representation of PED_UNI
        """

        self.gaussian = PED_GAUSS(conn, module_id, port_id, flow_index, impairment_type_index)
        """Gaussian distribution configuration.
        Representation of PED_GAUSS
        """

        self.poisson = PED_POISSON(conn, module_id, port_id, flow_index, impairment_type_index)
        """Poisson distribution configuration.
        Representation of PED_POISSON
        """

        self.gamma = PED_GAMMA(conn, module_id, port_id, flow_index, impairment_type_index)
        """Gamma distribution configuration.
        Representation of PED_GAMMA
        """

        self.custom = PED_CUST(conn, module_id, port_id, flow_index, impairment_type_index)
        """Associate a custom distribution to a flow and impairment type.
        Representation of PED_CUST
        """

        self.constant_delay = PED_CONST(conn, module_id, port_id, flow_index, impairment_type_index)
        """Constant Delay distribution configuration.
        Representation of PED_CONST
        """

        self.accumulate_and_burst = PED_ACCBURST(conn, module_id, port_id, flow_index, impairment_type_index)
        """Accumulate & Burst distribution configuration.
        Representation of PED_ACCBURST
        """

        self.step = PED_STEP(conn, module_id, port_id, flow_index, impairment_type_index)
        """Step distribution configuration.
        Representation of PED_STEP
        """

        self.fixed_burst = PED_FIXEDBURST(conn, module_id, port_id, flow_index, impairment_type_index)
        """Fixed Burst distribution configuration.
        Representation of PED_FIXEDBURST
        """



from __future__ import annotations
from dataclasses import dataclass
import typing
import functools

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpInt,
    XmpLong,
    XmpStr
)
from .enums import (
    OnOff,
    TPLDMode,
    CorruptionType,
    PolicerMode,
)


@register_command
@dataclass
class PE_FCSDROP:
    """
    The action on packets with FCS errors on a port.
    """

    code: typing.ClassVar[int] = 1601
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, the action on packets with FCS errors on a port"""

    class SetDataAttr(RequestBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, the action on packets with FCS errors on a port"""

    def get(self) -> Token[GetDataAttr]:
        """Get the status of whether the action on packets with FCS errors on a port is enabled.

        :return: whether the action on packets with FCS errors on a port is enabled
        :rtype: PE_FCSDROP.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, on_off: OnOff) -> Token[None]:
        """Set the status of whether the action on packets with FCS errors on a port is enabled.

        :param on_off:  whether the action on packets with FCS errors on a port is enabled
        :type on_off: OnOff
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable the action on packets with FCS errors on a port.
    """

    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable the action on packets with FCS errors on a port.
    """


@register_command
@dataclass
class PE_TPLDMODE:
    """
    The action indicates the TPLD mode to be used per port.
    """

    code: typing.ClassVar[int] = 1602
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        mode: TPLDMode = field(XmpByte())
        """coded byte, indicating the TPLD mode."""

    class SetDataAttr(RequestBodyStruct):
        mode: TPLDMode = field(XmpByte())
        """coded byte, indicating the TPLD mode."""

    def get(self) -> Token[GetDataAttr]:
        """Get the TPLD mode of the port.

        :return: indicating the TPLD mode
        :rtype: PE_TPLDMODE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, mode: TPLDMode) -> Token[None]:
        """Set the TPLD mode of the port.

        :param mode: indicating the TPLD mode
        :type mode: TPLDMode
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, mode=mode))

    set_normal = functools.partialmethod(set, TPLDMode.NORMAL)
    """Set the TPLD mode to Normal.
    """

    set_micro = functools.partialmethod(set, TPLDMode.MICRO)
    """Set the TPLD mode to Micro.
    """


@register_command
@dataclass
class PE_COMMENT:
    """
    Flow description.
    """

    code: typing.ClassVar[int] = 1605
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        comment: str = field(XmpStr())
        """string, containing the description of the flow."""

    class SetDataAttr(RequestBodyStruct):
        comment: str = field(XmpStr())
        """string, containing the description of the flow."""

    def get(self) -> Token[GetDataAttr]:
        """Get the flow description.

        :return: the description of the flow
        :rtype: PE_COMMENT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))

    def set(self, comment: str) -> Token[None]:
        """Set the flow description.

        :param comment: the description of the flow
        :type comment: str
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex], comment=comment))


@register_command
@dataclass
class PE_INDICES:
    """
    Get the flow indices. Currently the number of flows is 8.
    """

    code: typing.ClassVar[int] = 1608
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        flow0_id: int = field(XmpInt())
        """integer, id of flow 0."""
        flow1_id: int = field(XmpInt())
        """integer, id of flow 1."""
        flow2_id: int = field(XmpInt())
        """integer, id of flow 2."""
        flow3_id: int = field(XmpInt())
        """integer, id of flow 3."""
        flow4_id: int = field(XmpInt())
        """integer, id of flow 4."""
        flow5_id: int = field(XmpInt())
        """integer, id of flow 5."""
        flow6_id: int = field(XmpInt())
        """integer, id of flow 6."""
        flow7_id: int = field(XmpInt())
        """integer, id of flow 7."""

    def get(self) -> Token[GetDataAttr]:
        """Get the flow indices of a port. Currently the number of flows is fixed to 8.

        :return: the flow indices of a port
        :rtype: PE_INDICES.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_LATENCYRANGE:
    """
    Retrieve minimum and maximum configurable latency per flow in nanoseconds.
    """

    code: typing.ClassVar[int] = 1646
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        min: int = field(XmpLong())
        """long integer, minimum configurable latency per flow in nanoseconds."""
        max: int = field(XmpLong())
        """long integer, maximum configurable latency per flow in nanoseconds."""

    def get(self) -> Token[GetDataAttr]:
        """Get the minimum and maximum configurable latency per flow in nanoseconds.

        :return: minimum and maximum configurable latency per flow in nanoseconds.
        :rtype: PE_LATENCYRANGE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_CORRUPT:
    """
    Configures impairment corruption type.

    .. note::

        IP / TCP / UDP corruption modes are not supported on default flow (0)

    """

    code: typing.ClassVar[int] = 1660
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        corruption_type: CorruptionType = field(XmpByte())
        """coded byte, specifying corruption type."""

    class SetDataAttr(RequestBodyStruct):
        corruption_type: CorruptionType = field(XmpByte())
        """coded byte, specifying corruption type."""

    def get(self) -> Token[GetDataAttr]:
        """Get the impairment corruption type of a flow.

        :return: the impairment corruption type of a flow.
        :rtype: PE_CORRUPT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))

    def set(self, corruption_type: CorruptionType) -> Token[None]:
        """Set the impairment corruption type.

        :param corruption_type: corruption type
        :type corruption_type: CorruptionType
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex], corruption_type=corruption_type))

    set_off = functools.partialmethod(set, CorruptionType.OFF)
    """Set the corruption type to OFF.
    """

    set_eth = functools.partialmethod(set, CorruptionType.ETH)
    """Set the corruption type to Ethernet.
    """

    set_ip = functools.partialmethod(set, CorruptionType.IP)
    """Set the corruption type to IP.
    """

    set_udp = functools.partialmethod(set, CorruptionType.UDP)
    """Set the corruption type to UDP.
    """

    set_tcp = functools.partialmethod(set, CorruptionType.TCP)
    """Set the corruption type to TCP.
    """

    set_ber = functools.partialmethod(set, CorruptionType.BER)
    """Set the corruption type to Bit Error Rate.
    """


@register_command
@dataclass
class PE_MISORDER:
    """
    Configures the misordering depth in number of packets.

    .. note::

        probability * (depth + 1) should be less than 1,000,000. (see PED_FIXED)

    """

    code: typing.ClassVar[int] = 1661
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        depth: int = field(XmpInt())
        """integer, specifies the misordering depth (Range 1 - 32). Default value. probability * (depth + 1) should be less than 1,000,000. (see PED_FIXED)"""

    class SetDataAttr(RequestBodyStruct):
        depth: int = field(XmpInt())
        """integer, specifies the misordering depth (Range 1 - 32). Default value. probability * (depth + 1) should be less than 1,000,000. (see PED_FIXED)"""

    def get(self) -> Token[GetDataAttr]:
        """Get the misordering depth in number of packets of a flow.

        :return: the misordering depth (Range 1 - 32). Default value.
        :rtype: PE_MISORDER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))

    def set(self, depth: int) -> Token[None]:
        """Set the misordering depth in number of packets of a flow. Note: probability [see
        PED_FIXED] * (depth + 1) should be less than 1,000,000.

        :param depth: the misordering depth (Range 1 - 32). Default value.
        :type depth: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex], depth=depth))


@register_command
@dataclass
class PE_BANDPOLICER:
    """
    Configures the bandwidth policer.
    """

    code: typing.ClassVar[int] = 1662
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, enables/disables policer. Note: PED_ENABLE is not supported for the policer."""
        mode: PolicerMode = field(XmpByte())
        """coded byte, sets policer mode."""
        cir: int = field(XmpInt())
        """integer, policer committed information rate in units of 100 kbps (range 0 to 1000000), default is 0."""
        cbs: int = field(XmpInt())
        """integer, policer committed burst burst in bytes (range 0 to 4194304), default is 0."""

    class SetDataAttr(RequestBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, enables/disables policer. Note: PED_ENABLE is not supported for the policer."""
        mode: PolicerMode = field(XmpByte())
        """coded byte, sets policer mode."""
        cir: int = field(XmpInt())
        """integer, policer committed information rate in units of 100 kbps (range 0 to 1000000), default is 0."""
        cbs: int = field(XmpInt())
        """integer, policer committed burst size in bytes (range 0 to 4194304), default is 0."""

    def get(self) -> Token[GetDataAttr]:
        """Get the bandwidth policer configuration.

        :return: enabled/disabled, policer mode, committed information rate, and committed burst size.
        :rtype: PE_BANDPOLICER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))

    def set(self, on_off: OnOff, mode: PolicerMode, cir: int, cbs: int) -> Token[None]:
        """Set the bandwidth policer configuration.

        :param on_off: enables/disables policer. Note: PED_ENABLE is not supported for the policer.
        :type on_off: OnOff
        :param mode: policer mode
        :type mode: PolicerMode
        :param cir: policer committed information rate in units of 100 kbps (range 0 to 1000000), default is 0.
        :type cir: int
        :param cbs: policer committed burst burst in bytes (range 0 to 4194304), default is 0.
        :type cbs: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex], on_off=on_off, mode=mode, cir=cir, cbs=cbs))


@register_command
@dataclass
class PE_BANDSHAPER:
    """
    Configures the bandwidth shaper. L1 (0) (Shaper performed at Layer 1 level. I.e. including
    the preamble and min interpacket gap) L2 (1) (Shaper performed at Layer 2 level. I.e. excluding the preamble and min interpacket gap) Default value: L2 (0)
    """

    code: typing.ClassVar[int] = 1663
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, enables/disables shaper."""
        mode: PolicerMode = field(XmpByte())
        """coded byte, sets shaper mode."""
        cir: int = field(XmpInt())
        """integer, shaper committed information rate in units of 100 kbps (range 0 to 1000000), default is 0."""
        cbs: int = field(XmpInt())
        """integer, shaper committed burst size in bytes (range 0 to 4194304), default is 0."""
        buffer_size: int = field(XmpInt())
        """integer, shaper buffer size in bytes (range 0 to 2097152), default is 0."""

    class SetDataAttr(RequestBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, enables/disables shaper."""
        mode: PolicerMode = field(XmpByte())
        """coded byte, sets shaper mode."""
        cir: int = field(XmpInt())
        """integer, shaper committed information rate in units of 100 kbps (range 0 to 1000000), default is 0."""
        cbs: int = field(XmpInt())
        """integer, shaper committed burst size in bytes (range 0 to 4194304), default is 0."""
        buffer_size: int = field(XmpInt())
        """integer, shaper buffer size in bytes (range 0 to 2097152), default is 0."""

    def get(self) -> Token[GetDataAttr]:
        """Get the bandwidth shaper configuration.

        :return: the bandwidth shaper configuration, including on/off, mode, CIR, CBS, and shaper buffer size.
        :rtype: PE_BANDSHAPER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))

    def set(self, on_off: OnOff, mode: PolicerMode, cir: int, cbs: int, buffer_size: int) -> Token[None]:
        """Set the bandwidth shaper configuration.

        :param on_off: enables/disables shaper
        :type on_off: OnOff
        :param mode: shaper mode
        :type mode: PolicerMode
        :param cir: shaper committed information rate in units of 100 kbps (range 0 to 1000000), default is 0.
        :type cir: int
        :param cbs: shaper committed burst size in bytes (range 0 to 4194304), default is 0.
        :type cbs: int
        :param buffer_size: shaper buffer size in bytes (range 0 to 2097152), default is 0.
        :type buffer_size: int
        """

        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._flow_xindex],
                on_off=on_off,
                mode=mode,
                cir=cir,
                cbs=cbs,
                buffer_size=buffer_size
            )
        )


@register_command
@dataclass
class PE_DROPTOTAL:
    """
    Obtains statistics concerning all the packets dropped between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1750
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_drop_count_total: int = field(XmpLong())
        """long integer, total number of packets dropped in all flows."""
        pkt_drop_count_programmed: int = field(XmpLong())
        """long integer, total number of packets dropped as programmed in all flows."""
        pkt_drop_count_bandwidth: int = field(XmpLong())
        """long integer, total number of packets dropped due to bandwidth control in all flows."""
        pkt_drop_count_other: int = field(XmpLong())
        """long integer, total number of packets dropped for other reasons in all flows."""
        pkt_drop_ratio_total: int = field(XmpLong())
        """long integer, ratio of number of packets dropped in all flows, expressed in ppm."""
        pkt_drop_ratio_programmed: int = field(XmpLong())
        """long integer, ratio of number of packets dropped as programmed in all flows, expressed in ppm."""
        pkt_drop_ratio_bandwidth: int = field(XmpLong())
        """long integer, ratio of number of packets dropped due to bandwidth control in all flows, expressed in ppm."""
        pkt_drop_ratio_other: int = field(XmpLong())
        """long integer, ratio of number of packets dropped for other reasons in all flows, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets dropped between this receive port and its partner TX port.

        :return:
            total number of packets dropped in all flows,
            total number of packets dropped as programmed in all flows,
            total number of packets dropped due to bandwidth control in all flows,
            total number of packets dropped for other reasons in all flows,
            ratio of number of packets dropped in all flows, expressed in ppm,
            ratio of number of packets dropped as programmed in all flows, expressed in ppm,
            ratio of number of packets dropped due to bandwidth control in all flows, expressed in ppm,
            ratio of number of packets dropped for other reasons in all flows, expressed in ppm.
        :rtype: PE_DROPTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_LATENCYTOTAL:
    """
    Obtains statistics concerning all the packets delayed this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1751
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets delayed in all flows."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets delayed in all flows, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets delayed this receive port and its
        partner TX port.

        :return: number of packets delayed in all flows, and ratio of number of packets delayed in all flows, expressed in ppm.
        :rtype: PE_LATENCYTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_DUPTOTAL:
    """
    Obtains statistics concerning all the packets duplicated between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1752
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets duplicated in all flows."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets duplicated in all flows, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets duplicated between this receive
        port and its partner TX port.

        :return: number of packets duplicated in all flows, ratio of number of packets duplicated in all flows, expressed in ppm.
        :rtype: PE_DUPTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_MISTOTAL:
    """
    Obtains statistics concerning all the packets mis-ordered between this receive
    port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1753
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets mis-ordered in all flows."""
        ratio: int = field(XmpLong())
        """long integer, number of packets mis-ordered in all flows, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets mis-ordered between this receive port and its partner TX port.

        :return: number of packets mis-ordered in all flows, number of packets mis-ordered in all flows, expressed in ppm.
        :rtype: PE_MISTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_CORTOTAL:
    """
    Obtains statistics concerning all the packets corrupted on between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1754
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        total_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets corrupted in all flows."""
        fcs_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with Ethernet FCS corrupted in all flows."""
        ip_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with IP header checksum corrupted in all flows."""
        udp_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with UDP checksum corrupted in all flows."""
        tcp_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with TCP checksum corrupted in all flows."""
        total_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets corrupted in all flows, expressed in ppm."""
        fcs_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with Ethernet FCS corrupted in all flows, expressed in ppm."""
        ip_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with IP Header checksum corrupted in all flows, expressed in ppm."""
        udp_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with UDP checksum corrupted in all flows, expressed in ppm."""
        tcp_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with TCP checksum corrupted in all flows, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets corrupted on between this receive port and its partner TX port.

        :return: number of packets corrupted in all flows;
            number of packets with Ethernet FCS corrupted in all flows;
            number of packets with IP header checksum corrupted in all flows;
            number of packets with UDP checksum corrupted in all flows;
            number of packets with TCP checksum corrupted in all flows;
            ratio of number of packets corrupted in all flows, expressed in ppm;
            ratio of number of packets with Ethernet FCS corrupted in all flows expressed in ppm;
            ratio of number of packets with IP Header checksum corrupted in all flows, expressed in ppm;
            ratio of number of packets with UDP checksum corrupted in all flows, expressed in ppm;
            ratio of number of packets with TCP checksum corrupted in all flows, expressed in ppm
        :rtype: PE_CORTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_JITTERTOTAL:
    """
    Obtains statistics concerning all the packets jittered between this receive port
    and its partner TX port.
    """

    code: typing.ClassVar[int] = 1755
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets jittered in all flows."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets jittered in all flows expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets jittered between this receive port
        and its partner TX port.

        :return: number of packets jittered in all flows, ratio of number of packets jittered in all flows expressed in ppm.
        :rtype: PE_JITTERTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_CLEAR:
    """
    Clear all the impairment (duplicate, drop, mis-ordered, corrupted, latency and
    jitter) statistics for a Chimera port and flows on the port. The byte and packet
    counts will restart at zero.
    """

    code: typing.ClassVar[int] = 1756
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Clear all the impairment (duplicate, drop, mis-ordered, corrupted, latency and
        jitter) statistics for a Chimera port and flows on the port. The byte and packet
        counts will restart at zero.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PE_FLOWDROPTOTAL:
    """
    Obtains statistics concerning all the packets dropped in a flow between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1770
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_drop_count_total: int = field(XmpLong())
        """long integer, total number of packets dropped for the flow."""
        pkt_drop_count_programmed: int = field(XmpLong())
        """long integer, total number of packets dropped as programmed for the flow."""
        pkt_drop_count_bandwidth: int = field(XmpLong())
        """long integer, total number of packets dropped due to bandwidth control for the flow."""
        pkt_drop_count_other: int = field(XmpLong())
        """long integer, total number of packets dropped for other reasons for the flow."""
        pkt_drop_ratio_total: int = field(XmpLong())
        """long integer, ratio of number of packets dropped for the flow, expressed in ppm."""
        pkt_drop_ratio_programmed: int = field(XmpLong())
        """long integer, ratio of number of packets dropped as programmed for the flow, expressed in ppm."""
        pkt_drop_ratio_bandwidth: int = field(XmpLong())
        """long integer, ratio of number of packets dropped due to bandwidth control for the flow, expressed in ppm."""
        pkt_drop_ratio_other: int = field(XmpLong())
        """long integer, ratio of number of packets dropped for other reasons for the flow, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets dropped in a flow between this receive port and its partner TX port.

        :return:
            total number of packets dropped for the flow,
            total number of packets dropped as programmed for the flow,
            total number of packets dropped due to bandwidth control for the flow,
            total number of packets dropped for other reasons for the flow,
            ratio of number of packets dropped for the flow, expressed in ppm,
            ratio of number of packets dropped as programmed for the flow, expressed in ppm,
            ratio of number of packets dropped due to bandwidth control for the flow, expressed in ppm,
            ratio of number of packets dropped for other reasons for the flow, expressed in ppm.
        :rtype: PE_FLOWDROPTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_FLOWLATENCYTOTAL:
    """
    Obtains statistics concerning all the packets delayed between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1771
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets delayed in the flow."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets delayed in the flow, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets delayed between this receive port and its partner TX port.

        :return: number of packets delayed in the flow, ratio of number of packets delayed in the flow, expressed in ppm.
        :rtype: PE_FLOWLATENCYTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_FLOWDUPTOTAL:
    """
    Obtains statistics concerning all the packets duplicated in a flow between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1772
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets duplicated for the flow."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets duplicated for the flow - expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets duplicated in a flow between this receive port and its partner TX port.

        :return: number of packets duplicated for the flow, ratio of number of packets duplicated for the flow - expressed in ppm.
        :rtype: PE_FLOWDUPTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_FLOWMISTOTAL:
    """
    Obtains statistics concerning all the packets mis-ordered in a flow between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1773
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets mis-ordered for the flow."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets mis-ordered in a flow between this receive port and its partner TX port.

        :return: number of packets mis-ordered for the flow, ratio of number of packets, expressed in ppm.
        :rtype: PE_FLOWMISTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_FLOWCORTOTAL:
    """
    Obtains statistics concerning all the packets corrupted in a flow between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1774
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        total_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets corrupted for the flow."""
        fcs_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with Ethernet FCS corrupted for the flow."""
        ip_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with IP header checksum corrupted for the flow."""
        udp_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with UDP checksum corrupted for the flow."""
        tcp_corrupted_pkt_count: int = field(XmpLong())
        """long integer, number of packets with TCP checksum corrupted for the flow."""
        total_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets corrupted for the flow expressed in ppm."""
        fcs_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with Ethernet FCS corrupted for the flow expressed in ppm."""
        ip_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with IP Header checksum corrupted for the flow expressed in ppm."""
        udp_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with UDP checksum corrupted for the flow expressed in ppm."""
        tcp_corrupted_pkt_ratio: int = field(XmpLong())
        """long integer, ratio of number of packets with TCP checksum corrupted for the flow expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets corrupted in a flow between this receive port and its partner TX port.

        :return:
            number of packets corrupted for the flow,
            number of packets with Ethernet FCS corrupted for the flow,
            number of packets with IP header checksum corrupted for the flow,
            number of packets with UDP checksum corrupted for the flow,
            number of packets with TCP checksum corrupted for the flow,
            ratio of number of packets corrupted for the flow expressed in ppm,
            ratio of number of packets with Ethernet FCS corrupted for the flow expressed in ppm,
            ratio of number of packets with IP Header checksum corrupted for the flow expressed in ppm,
            ratio of number of packets with UDP checksum corrupted for the flow expressed in ppm,
            ratio of number of packets with TCP checksum corrupted for the flow expressed in ppm.
        :rtype: PE_FLOWCORTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_FLOWJITTERTOTAL:
    """
    Obtains statistics concerning all the packets jittered in a flow between this receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1775
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pkt_count: int = field(XmpLong())
        """long integer, number of packets jittered in the flow."""
        ratio: int = field(XmpLong())
        """long integer, ratio of number of packets jittered in the flow, expressed in ppm."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets jittered in a flow between this receive port and its partner TX port.

        :return: number of packets jittered in the flow, ratio of number of packets jittered in the flow, expressed in ppm
        :rtype: PE_FLOWJITTERTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PE_FLOWCLEAR:
    """
    Clear all the impairment (duplicate, drop, mis-ordered, corrupted, latency and
    jitter) statistics on a particular flow on the port. The byte and packet counts
    will restart at zero.
    """

    code: typing.ClassVar[int] = 1776
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Clear all the impairment (duplicate, drop, mis-ordered, corrupted, latency and
        jitter) statistics on a particular flow on the port. The byte and packet counts
        will restart at zero.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))



from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    C_TKLICFILE,
    C_TKLICSTATE,
    C_TKSTATUS,
    C_TKSTATUSEXT,
    C_TKSVCSTATE,
    C_TKGPSSTATE,
    C_TKCONFIG,
)


class TimeKeeper:
    """TimeKeeper of Valkyrie."""

    def __init__(self, conn: "itf.IConnection") -> None:
        self.license_file = C_TKLICFILE(conn)
        """TimeKeeper license file content.
        Representation of C_TKLICFILE
        """

        self.license_state = C_TKLICSTATE(conn)
        """State of TimeKeeper license file content.
        Representation of C_TKLICSTATE
        """

        self.status = C_TKSTATUS(conn)
        """Version and status of TimeKeeper.
        Representation of C_TKSTATUS
        """

        self.status_extended = C_TKSTATUSEXT(conn)
        """Version and status of TimeKeeper (extended).
        Representation of C_TKSTATUSEXT
        """

        self.svc_state = C_TKSVCSTATE(conn)
        """TimeKeeper service state.
        Representation of C_TKSVCSTATE
        """

        self.gps_state = C_TKGPSSTATE(conn)
        """TimeKeeper GPS state.
        Representation of C_TKGPSSTATE
        """

        self.config_file = C_TKCONFIG(conn)
        """TimeKeeper configuration file content.
        Representation of C_TKCONFIG
        """



#: Available enums used by commands and server response status.

from .internals.commands.enums import (
    AlgorithmMethod,
    AlignLockStatus,
    ApplicationLayerBehavior,
    AutoNegFECOption,
    AutoNegFECStatus,
    AutoNegFECType,
    AutoNegMode,
    AutoNegStatus,
    AutoNegTecAbility,
    AutoOrManual,
    BRRMode,
    CaptureSize,
    ChassisSessionType,
    ChassisShutdownAction,
    CongestionType,
    CorruptionType,
    CustomDefaultCommand,
    CustomDefaultScope,
    EmbedIP,
    ErrorOnOff,
    FaultSignaling,
    FECMode,
    FilterMode,
    FilterType,
    FilterUse,
    FilterVlanType,
    HasDemo,
    HeaderLockStatus,
    IGMPVersion,
    ImpairmentLatencyMode,
    ImpairmentTypeIndex,
    InfiniteOrFinite,
    InfoAction,
    InjectErrorType,
    IsEnabled,
    IsOnline,
    IsPermanent,
    IsPresent,
    IsValid,
    LinkTrainFrameLock,
    L2PlusPresent,
    L3Present,
    L47IPVersion,
    L47PortSpeed,
    L47PortState,
    L47ProtocolType,
    L47TrafficState,
    LatencyMode,
    LatencyTypeCustomDist,
    LengthCheckType,
    LengthType,
    LicenseSpeed,
    LifecycleMode,
    LinkState,
    LinkTrainingFailureType,
    LinkTrainingInitCondition,
    LinkTrainingMode,
    LinkTrainingStatus,
    LinkTrainingStatusMode,
    LocalFaultStatus,
    LoopbackMode,
    LoopBandwidth,
    MDIXMode,
    MediaCFPState,
    MediaCFPType,
    MediaConfigurationType,
    ModifierAction,
    MSSType,
    MulticastExtOperation,
    MulticastHeaderFormat,
    MulticastOperation,
    NRZPreset,
    OnOff,
    OnOffDefault,
    OnOffWithSuppress,
    PacketDetailSelection,
    PacketType,
    PAM4FrameSize,
    PauseMode,
    PayloadGenerationMethod,
    PayloadMode,
    PayloadType,
    PFCMode,
    PHYSignalStatus,
    PolicerMode,
    PortSpeedMode,
    PPMSweepMode,
    PPMSweepStatus,
    PRBSInsertedType,
    PRBSInvertState,
    PRBSLockStatus,
    PRBSOnOff,
    PRBSPattern,
    PRBSPolynomial,
    PRBSStatisticsMode,
    ProtocolOption,
    ReconciliationSublayerSupport,
    RemoteFaultStatus,
    ReplayParserState,
    ReplaySchedulingMode,
    ReplaySyncBasedOn,
    ReservedAction,
    ReservedStatus,
    ResourceAllocationMode,
    RESTControlAction,
    Role,
    RTOType,
    RXCState,
    RXHState,
    SerdesStatus,
    ServiceStatus,
    ShadowWorkingSelection,
    SMAInputFunction,
    SMAOutputFunction,
    SMAStatus,
    SourceType,
    StartOrStop,
    StartTrigger,
    StopTrigger,
    SyncStatus,
    SystemUpdateStatus,
    TimeKeeperLicenseError,
    TimeKeeperLicenseFileState,
    TimeKeeperLicenseType,
    TimeKeeperServiceAction,
    TimeKeeperServiceStatus,
    TimeoutMode,
    Timescale,
    TimingSource,
    TLSVersion,
    TPLDMode,
    TrafficEngine,
    TrafficError,
    TrafficOnOff,
    TrafficScenario,
    TSNConfigProfile,
    TSNDeviationMode,
    TSNHistogramSource,
    TSNPortRole,
    TSNStatisticsTypes,
    TSNTimeSource,
    TXClockSource,
    TXClockStatus,
    TXCState,
    TXHState,
    TXMode,
    UpdateState,
    WhoClose,
    YesNo,
    Layer1ConfigType,
    LinkTrainAlgorithm,
    LinkTrainCmd,
    LinkTrainAnnounce,
    Layer1LogType,
    LinkTrainCmdFlags,
    LinkTrainCmdResults,
    LinkTrainCoeffs,
    LinkTrainEncoding,
    LinkTrainPresets,
    AnLtLogControl,
    RxEqExtCap,
    RxEqExtCapMode,
    PreCodingMode,
    GrayCodingMode,
    Endianness,
    StreamOption,
    DhcpState,
    DhcpVlanState,
    VlanType,
    ChassisModelNumber,
    ChassisModelName,
    ModuleModelName,
    FreyaAutonegMode,
    FreyaLinkTrainingMode,
    FreyaTecAbility,
    FreyaFECAbility,
    FreyaPauseAbility,
    FreyaTechAbilityHCDStatus,
    FreyaOutOfSyncPreset,
    Layer1Control,
    FreyaPCSVariant,
    FreyaTecAbilityHCD,
    Layer1Opcode,
)

__all__ = (
    "AlgorithmMethod",
    "AlignLockStatus",
    "ApplicationLayerBehavior",
    "AutoNegFECOption",
    "AutoNegFECStatus",
    "AutoNegFECType",
    "AutoNegMode",
    "AutoNegStatus",
    "AutoNegTecAbility",
    "AutoOrManual",
    "BRRMode",
    "CaptureSize",
    "ChassisSessionType",
    "ChassisShutdownAction",
    "CongestionType",
    "CorruptionType",
    "CustomDefaultCommand",
    "CustomDefaultScope",
    "EmbedIP",
    "ErrorOnOff",
    "FaultSignaling",
    "FECMode",
    "FilterMode",
    "FilterType",
    "FilterUse",
    "FilterVlanType",
    "HasDemo",
    "HeaderLockStatus",
    "IGMPVersion",
    "ImpairmentLatencyMode",
    "ImpairmentTypeIndex",
    "InfiniteOrFinite",
    "InfoAction",
    "InjectErrorType",
    "IsEnabled",
    "IsOnline",
    "IsPermanent",
    "IsPresent",
    "IsValid",
    "LinkTrainFrameLock",
    "L2PlusPresent",
    "L3Present",
    "L47IPVersion",
    "L47PortSpeed",
    "L47PortState",
    "L47ProtocolType",
    "L47TrafficState",
    "LatencyMode",
    "LatencyTypeCustomDist",
    "LengthCheckType",
    "LengthType",
    "LicenseSpeed",
    "LifecycleMode",
    "LinkState",
    "LinkTrainingFailureType",
    "LinkTrainingInitCondition",
    "LinkTrainingMode",
    "LinkTrainingStatus",
    "LinkTrainingStatusMode",
    "LocalFaultStatus",
    "LoopbackMode",
    "LoopBandwidth",
    "MDIXMode",
    "MediaCFPState",
    "MediaCFPType",
    "MediaConfigurationType",
    "ModifierAction",
    "MSSType",
    "MulticastExtOperation",
    "MulticastHeaderFormat",
    "MulticastOperation",
    "NRZPreset",
    "OnOff",
    "OnOffDefault",
    "OnOffWithSuppress",
    "PacketDetailSelection",
    "PacketType",
    "PAM4FrameSize",
    "PauseMode",
    "PayloadGenerationMethod",
    "PayloadMode",
    "PayloadType",
    "PFCMode",
    "PHYSignalStatus",
    "PolicerMode",
    "PortSpeedMode",
    "PPMSweepMode",
    "PPMSweepStatus",
    "PRBSInsertedType",
    "PRBSInvertState",
    "PRBSLockStatus",
    "PRBSOnOff",
    "PRBSPattern",
    "PRBSPolynomial",
    "PRBSStatisticsMode",
    "ProtocolOption",
    "ReconciliationSublayerSupport",
    "RemoteFaultStatus",
    "ReplayParserState",
    "ReplaySchedulingMode",
    "ReplaySyncBasedOn",
    "ReservedAction",
    "ReservedStatus",
    "ResourceAllocationMode",
    "RESTControlAction",
    "Role",
    "RTOType",
    "RXCState",
    "RXHState",
    "SerdesStatus",
    "ServiceStatus",
    "ShadowWorkingSelection",
    "SMAInputFunction",
    "SMAOutputFunction",
    "SMAStatus",
    "SourceType",
    "StartOrStop",
    "StartTrigger",
    "StopTrigger",
    "SyncStatus",
    "SystemUpdateStatus",
    "TimeKeeperLicenseError",
    "TimeKeeperLicenseFileState",
    "TimeKeeperLicenseType",
    "TimeKeeperServiceAction",
    "TimeKeeperServiceStatus",
    "TimeoutMode",
    "Timescale",
    "TimingSource",
    "TLSVersion",
    "TPLDMode",
    "TrafficEngine",
    "TrafficError",
    "TrafficOnOff",
    "TrafficScenario",
    "TSNConfigProfile",
    "TSNDeviationMode",
    "TSNHistogramSource",
    "TSNPortRole",
    "TSNStatisticsTypes",
    "TSNTimeSource",
    "TXClockSource",
    "TXClockStatus",
    "TXCState",
    "TXHState",
    "TXMode",
    "UpdateState",
    "WhoClose",
    "YesNo",
    "Layer1ConfigType",
    "LinkTrainAlgorithm",
    "LinkTrainCmd",
    "LinkTrainAnnounce",
    "Layer1LogType",
    "LinkTrainCmdFlags",
    "LinkTrainCmdResults",
    "LinkTrainCoeffs",
    "LinkTrainEncoding",
    "LinkTrainPresets",
    "AnLtLogControl",
    "RxEqExtCap",
    "RxEqExtCapMode",
    "PreCodingMode",
    "GrayCodingMode",
    "Endianness",
    "StreamOption",
    "DhcpState",
    "DhcpVlanState",
    "VlanType",
    "ChassisModelNumber",
    "ChassisModelName",
    "ModuleModelName",
    "FreyaAutonegMode",
    "FreyaLinkTrainingMode",
    "FreyaTecAbility",
    "FreyaFECAbility",
    "FreyaPauseAbility",
    "FreyaTechAbilityHCDStatus",
    "FreyaOutOfSyncPreset",
    "Layer1Control",
    "FreyaPCSVariant",
    "FreyaTecAbilityHCD",
    "Layer1Opcode",
)



from __future__ import annotations


import timeit
import pytest
import pstats
import cProfile

import asyncio
import os
import sys
import logging
from typing import Coroutine, List
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from loguru import logger  # noqa: E402
from xoa_driver.testers import L23Tester  # noqa: E402
from xoa_driver.lli import establish_connection  # noqa: E402
from xoa_driver.lli import commands  # noqa: E402
from xoa_driver.lli import TransportationHandler  # noqa: E402
from xoa_driver.utils import apply, apply_iter  # noqa: E402

# TODO: <GET> response contain more fields then driver can parse
# TODO: <GET> response contains less fields then driver can parse
# TODO: <SET> request require more fields then server expecting
# TODO: <SET> request require less fields then server expects


MULTIPLIER = 10_000

@pytest.mark.asyncio
async def test_lli() -> None:
    logging.basicConfig(
        format='%(relativeCreated)5d %(name)-15s %(levelname)-8s %(message)s',
        level=logging.DEBUG
    )
    logger_ = logging.getLogger(__file__)
    ctx = TransportationHandler()
    await establish_connection(ctx, "87.61.110.118")

    # # print("Is connected", ctx.is_connected)
    # with cProfile.Profile() as pr:
    *_, cc, mc, pc = await apply(
        commands.C_LOGON(ctx).set("xena"),
        commands.C_OWNER(ctx).set("xoa" * 10),
        # commands.C_OWNER(ctx).get(),
        # commands.C_SERIALNO(ctx).get(),
        # commands.C_MODEL(ctx).get()
        # commands.C_CAPABILITIES(ctx).get(),
        # commands.M_CAPABILITIES(ctx, 1).get(),
        # commands.P_CAPABILITIES(ctx, 1, 1).get(),
        # commands.P_ARPRXTABLE(ctx, 3, 1).get()
        commands.PED_SCHEDULE(ctx, 2, 3, 1, 0).get()
    )
    print(repr(pc), pc)
    # a = await commands.C_INDICES(ctx).get()
    # print((await commands.C_STATSESSION(ctx, a.session_ids[-1]).get()).to_bytes())
    # port_counts = (await commands.C_PORTCOUNTS(ctx).get()).port_counts

    # indices = (
    #     slot_id
    #     for slot_id, p_count in enumerate(port_counts)
    #     if p_count > 0
    # )
    # for mid in indices:
    #     r = await commands.M_REVISION(ctx, mid).get()
    #     print(
    #         r.revision,
    #         r.to_bytes(),
    #         r.nbytes(),
    #     )
    # tasks = (commands.P_CAPABILITIES(ctx, 1, 1).get() for _ in range(MULTIPLIER))
    # async for resp in apply_iter(*tasks):
    #     resp.tx_eq_tap_max_val
    # stats = pstats.Stats(pr)
    # stats.sort_stats(pstats.SortKey.TIME)
    # print(MULTIPLIER)
    # stats.print_stats(20)
    ctx.close()

    # port = PortL23()
    # if isinstance(port, PortL23) and port.is_capable_of(ANLT, PCS_PMA)

    # Testsuite 2544:
    # required_port_type: L23Port
    # required_functionalities:
    # ANLT functionalities:
    # required_port_type: L23Port,
    # required_functionalities:  ANLT


def run(method: Coroutine) -> None:
    import platform
    if platform.system() == 'Windows':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.run(method)


if __name__ == "__main__":
    run(test_lli())
    result = timeit.timeit(
        "run(test_lli())",
        setup="from __main__ import run, test_lli",
        number=1
    )
    print(
        f"Time pass: {result}, Packets Sent: {MULTIPLIER}, PPS: {MULTIPLIER / result}"
    )



from enum import IntEnum, auto


class IndexEvents(IntEnum):
    DEL = auto()


class IndicesObserver:
    def __init__(self) -> None:
        self.__subscriptions = {}

    def subscribe(self, evt, func) -> None:
        assert evt not in self.__subscriptions, f"Event <{evt}> is already registered."
        self.__subscriptions[evt] = func

    def notify(self, evt, index_inst) -> None:
        self.__subscriptions[evt](index_inst)



from __future__ import annotations

import timeit
import pstats
import cProfile
import pytest
import asyncio
import os
import sys
import logging
from typing import Coroutine, List

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from loguru import logger  # noqa: E402
from xoa_driver.testers import L23Tester  # noqa: E402



@pytest.mark.asyncio
async def test_hli() -> None:
    # logger.add(
    #     "./app.log",
    #     rotation="10 MB",
    #     retention=5,
    #     enqueue=True,
    #     compression="zip",
    #     level=0,
    #     serialize=True
    # )

    async with L23Tester("192.168.1.198", "ACO", enable_logging=True, custom_logger=logger) as tester:
        module = tester.modules.obtain(1)
        port1 = module.ports.obtain(0)
        port2 = module.ports.obtain(1)
        print(port1, port2)
        # print(port1.info.reservation)
        # print(port1.info.reserved_by)
        # await port1.reservation.set_reserve()



def run(method: Coroutine) -> None:
    import platform

    if platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.run(method)


if __name__ == "__main__":
    run(test_hli())
    # result = timeit.timeit(
    #     "run(main())",
    #     setup="from __main__ import run, main",
    #     number=10
    # )
    # print(result)



from __future__ import annotations
from dataclasses import dataclass
import typing

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpHex,
    Hex,
)


@register_command
@dataclass
class P4E_ASSIGN:
    """
    Advanced mode only: Assign previously reserved PEs to a port.
    """

    code: typing.ClassVar[int] = 675
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        mask: Hex = field(XmpHex(size=8))
        """eight hex bytes, a bitmask specifying which PEs should be assigned to this port"""

    class SetDataAttr(RequestBodyStruct):
        mask: Hex = field(XmpHex(size=8))
        """eight hex bytes, a bitmask specifying which PEs should be assigned to this port"""

    def get(self) -> Token[GetDataAttr]:
        """Get the assigned PEs on the port.

        :return: previously reserved PEs
        :rtype: P4E_ASSIGN.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, mask: Hex) -> Token[None]:
        """Assign the previously reserved PEs to the port.

        :param mask: a bitmask specifying which PEs should be assigned to this port
        :type mask: str
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, mask=mask))


@register_command
@dataclass
class P4E_AVAILABLE:
    """
    Simple mode only: Report the number of PEs available for allocation.
    """

    code: typing.ClassVar[int] = 676
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        available_pe_count: int = field(XmpByte())
        """byte, total number of PEs that can be allocated to the port - including the PEs already allocated to the port."""

    def get(self) -> Token[GetDataAttr]:
        """Get the number of PEs available for allocation.

        :return: the number of PEs available for allocation.
        :rtype: P4E_AVAILABLE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4E_ALLOCATE:
    """
    Simple mode only: Allocate a number of PEs to this port.
    """

    code: typing.ClassVar[int] = 677
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        pe_count_alloc: int = field(XmpByte())
        """byte, the total number of PEs to allocate to this port - including the PEs already allocated to the port."""

    class SetDataAttr(RequestBodyStruct):
        pe_count_alloc: int = field(XmpByte())
        """byte, the total number of PEs to allocate to this port - including the PEs already allocated to the port."""

    def get(self) -> Token[GetDataAttr]:
        """Get the number of PEs allocated to this port.

        :return: the total number of PEs to allocate to this port - including the PEs already allocated to the port.
        :rtype: P4E_ALLOCATE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, pe_count_alloc: int) -> Token[None]:
        """Allocate a number of PEs to this port.

        :param pe_count_alloc: the total number of PEs to allocate to this port - including the PEs already allocated to the port.
        :type pe_count_alloc: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, pe_count_alloc=pe_count_alloc))


@register_command
@dataclass
class P4E_ALLOCATION_INFO:
    """
    Display information about which PEs that are available for allocation/assignment
    and which are currently allocated/assigned to this port.
    """

    code: typing.ClassVar[int] = 678
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        available: Hex = field(XmpHex(size=8))
        """eight hex bytes (64-bit) mask of available PEs"""
        allocated: Hex = field(XmpHex(size=8))
        """eight hex bytes (64-bit) mask of PEs assigned to this port"""

    def get(self) -> Token[GetDataAttr]:
        """Get PEs allocation information.

        :return: PEs allocation information
        :rtype: P4E_ALLOCATION_INFO.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))



from typing import TYPE_CHECKING
from xoa_driver.internals.commands import (
    PS_INSERTFCS,
    PS_INJECTFCSERR,
)
from .base_stream import (
    BaseStreamIdx,
    SInjectError
)

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
    from xoa_driver.internals.utils.indices import observer as idx_obs


class GSInjectError(SInjectError):
    """Error Injection configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, stream_idx: int) -> None:
        super().__init__(conn, module_id, port_id, stream_idx)
        self.frame_checksum = PS_INJECTFCSERR(conn, module_id, port_id, stream_idx)
        """Inject FCS error

        :type: PS_INJECTFCSERR
        """


class GenuineStreamIdx(BaseStreamIdx):
    """Genuine L23 Stream Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)
        self.insert_packets_checksum = PS_INSERTFCS(conn, *kind)
        """Whether a valid frame checksum is added to the packets of a stream.

        :type:  PS_INSERTFCS
        """

        self.inject_err = GSInjectError(conn, *kind)
        """Error injection configuration

        :type: GSInjectError
        """



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    P4G_UDP_PACKET_SIZE_TYPE,
    P4G_UDP_PACKET_SIZE_MINMAX,
    P4G_UDP_PACKET_SIZE_VALUE,
    P4G_UDP_STATE_CURRENT,
    P4G_UDP_STATE_TOTAL,
    P4G_UDP_STATE_RATE,
    P4G_UDP_RX_PAYLOAD_COUNTERS,
    P4G_UDP_TX_PAYLOAD_COUNTERS,
    P4G_UDP_RX_BYTES_HIST,
    P4G_UDP_TX_BYTES_HIST,
    P4G_UDP_RX_PACKET_COUNTERS,
    P4G_UDP_TX_PACKET_COUNTERS,
)


class GHistogramUdp:
    """UDP histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx_bytes = P4G_UDP_RX_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_RX_BYTES_HIST
        """
        self.tx_bytes = P4G_UDP_TX_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_TX_BYTES_HIST
        """


class GPayloadCountersUdp:
    """UDP payload counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx = P4G_UDP_RX_PAYLOAD_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_RX_PAYLOAD_COUNTERS
        """
        self.tx = P4G_UDP_TX_PAYLOAD_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_TX_PAYLOAD_COUNTERS
        """


class GPacketCountersUdp:
    """UDP packet counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx = P4G_UDP_RX_PACKET_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_RX_PACKET_COUNTERS
        """
        self.tx = P4G_UDP_TX_PACKET_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_TX_PACKET_COUNTERS
        """


class GCountersUdp:
    """UDP counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.payload = GPayloadCountersUdp(conn, module_id, port_id, group_idx)
        """UDP payload counters"""
        self.packet = GPacketCountersUdp(conn, module_id, port_id, group_idx)
        """UDP packet counters"""
        self.state = GStateCountersUdp(conn, module_id, port_id, group_idx)
        """UDP state counters"""


class GPacketSizeUdp:
    """UDP packet size"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.type = P4G_UDP_PACKET_SIZE_TYPE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_PACKET_SIZE_TYPE
        """
        self.range_limits = P4G_UDP_PACKET_SIZE_MINMAX(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_PACKET_SIZE_MINMAX
        """
        self.value = P4G_UDP_PACKET_SIZE_VALUE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_PACKET_SIZE_VALUE
        """


class GStateCountersUdp:
    """UDP state counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.current = P4G_UDP_STATE_CURRENT(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_STATE_CURRENT
        """
        self.total = P4G_UDP_STATE_TOTAL(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_STATE_TOTAL
        """
        self.rate = P4G_UDP_STATE_RATE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_UDP_STATE_RATE
        """


class GUdp:
    """UDP configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.packet_size = GPacketSizeUdp(conn, module_id, port_id, group_idx)
        """UDP packet size"""
        self.counters = GCountersUdp(conn, module_id, port_id, group_idx)
        """UDP counters"""
        self.histogram = GHistogramUdp(conn, module_id, port_id, group_idx)
        """UDP histogram"""



from typing import TYPE_CHECKING
from xoa_driver.internals.commands import PF_STRING

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
    from xoa_driver.internals.utils.indices import observer as idx_obs

from .base_filter import BaseFilterIdx


class GenuineFilterIdx(BaseFilterIdx):
    """Genuine L23 Filter Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)
        self.string = PF_STRING(conn, *kind)
        """String representation of the filter.

        :type:  PF_STRING
        """



from typing import (
    TYPE_CHECKING,
    Tuple,
)
from typing_extensions import Self
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    PP_ALARMS_ERRORS,
    PP_TXLANECONFIG,
    PP_TXLANEINJECT,
    PP_TXPRBSCONFIG,
    PP_TXERRORRATE,
    PP_TXINJECTONE,
    PP_RXTOTALSTATS,
    PP_RXFECSTATS,
    PP_RXLANELOCK,
    PP_RXLANESTATUS,
    PP_RXLANEERRORS,
    PP_RXPRBSSTATUS,
    PP_RXCLEAR,
    PP_RXLASERPOWER,
    PP_TXLASERPOWER,
    PP_EYEMEASURE,
    PP_EYERESOLUTION,
    PP_EYEREAD,
    PP_EYEINFO,
    PP_PHYTXEQ,
    PP_PHYRETUNE,
    PP_PHYAUTOTUNE,
    PP_EYEBER,
    PP_PHYAUTONEG,
    PP_TXPRBSTYPE,
    PP_RXPRBSTYPE,
    # PP_FECMODE, # moved to all genuine ports
    PP_EYEDWELLBITS,
    PP_PHYSIGNALSTATUS,
    PP_PRBSTYPE,
    PP_PHYSETTINGS,
    PP_PHYRXEQ,
    PP_PRECODING,
    PP_GRAYCODING,
    PP_PRECODINGSTATUS,
)


class PcsPmaAlarms:
    """L23 high-speed port PCS/PMA alarms"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.errors = PP_ALARMS_ERRORS(conn, module_id, port_id)
        """Error count of each alarm on a L23 high-speed port.
        Representation of PP_ALARMS_ERRORS
        """


class PcsPmaTransceiver:
    """L23 high-speed port PCS/PMA transceivers"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.tx_laser_power = PP_TXLASERPOWER(conn, module_id, port_id)
        """Power of TX laser.
        Representation of PP_TXLASERPOWER
        """

        self.rx_laser_power = PP_RXLASERPOWER(conn, module_id, port_id)
        """Power of RX laser.
        Representation of PP_RXLASERPOWER
        """


class PcsPmaRxLaneStatus:
    """L23 high-speed port PCS/PMA lane status"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, lane_idx: int) -> None:
        self.errors = PP_RXLANEERRORS(conn, module_id, port_id, lane_idx)
        """RX lane error statistics.
        Representation of PP_RXLANEERRORS
        """

        self.lock = PP_RXLANELOCK(conn, module_id, port_id, lane_idx)
        """RX lane lock.
        Representation of PP_RXLANELOCK
        """

        self.status = PP_RXLANESTATUS(conn, module_id, port_id, lane_idx)
        """RX lane status
        Representation of PP_RXLANESTATUS
        """


class PcsPmaTxErrorGeneration:
    """L23 high-speed port PCS/PMA TX error generation."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.error_rate = PP_TXERRORRATE(conn, module_id, port_id)
        """The rate of continuous bit-level error injection.
        Representation of PP_TXERRORRATE
        """

        self.inject_one = PP_TXINJECTONE(conn, module_id, port_id)
        """Inject a single bit-level error.
        Representation of PP_TXINJECTONE
        """


class PcsPmaRx:
    """L23 high-speed port PCS/PMA RX"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.total_status = PP_RXTOTALSTATS(conn, module_id, port_id)
        """RX FEC total counters.
        Representation of PP_RXTOTALSTATS
        """

        self.fec_status = PP_RXFECSTATS(conn, module_id, port_id)
        """RX FEC statistics.
        Representation of PP_RXFECSTATS
        """

        self.clear = PP_RXCLEAR(conn, module_id, port_id)
        """Clear all the PCS/PMA receiver statistics.
        Representation of PP_RXCLEAR
        """


class PcsPmaPhy:
    """L23 high-speed port PCS/PMA PHY settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.auto_neg = PP_PHYAUTONEG(conn, module_id, port_id)
        """ Auto-negotiation settings of the PHY.
        Representation of PP_PHYAUTONEG
        """

        self.signal_status = PP_PHYSIGNALSTATUS(conn, module_id, port_id)
        """The PHY signal status.
        Representation of PP_PHYSIGNALSTATUS
        """

        self.settings = PP_PHYSETTINGS(conn, module_id, port_id)
        """Low-level PHY settings
        Representation of PP_PHYSETTINGS
        """


class Lane:
    """L23 high-speed port lane configuration and status."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, lane_idx: int) -> None:
        self.rx_status = PcsPmaRxLaneStatus(conn, module_id, port_id, lane_idx)
        """PCS/PMA RX lane status.
        """

        self.tx_error_inject = PP_TXLANEINJECT(conn, module_id, port_id, lane_idx)
        """Inject CAUI error into a TX lane.
        Representation of PP_TXLANEINJECT
        """

        self.tx_config = PP_TXLANECONFIG(conn, module_id, port_id, lane_idx)
        """TX lane configuration.
        Representation of PP_TXLANECONFIG
        """


class PcsPma:
    """L23 high-speed port PCS/PMA"""

    def __init__(self, conn: "itf.IConnection", port) -> None:
        self._conn = conn
        self.__port = port

        self.alarms = PcsPmaAlarms(conn, *port.kind)
        """PCS/PMA alarms"""

        self.transceiver = PcsPmaTransceiver(conn, *port.kind)
        """PCS/PMA transceiver"""

        self.error_gen = PcsPmaTxErrorGeneration(conn, *port.kind)
        """PCS/PMA error generation"""

        self.rx = PcsPmaRx(conn, *port.kind)
        """PCS/PMA RX"""

        self.phy = PcsPmaPhy(conn, *port.kind)
        """PCS/PMA PHY"""

        self.prbs_config = PRBSConfig(conn, *port.kind)
        """PCS/PMA PRBS configuration"""

        self.lanes: Tuple["Lane", ...] = tuple(
            Lane(self._conn, *self.__port.kind, lane_idx=idx)
            for idx in range(self.__port.info.capabilities.lane_count)
        )  # TODO: need to fix, currently port.info.capabilities must be none because lanes are created before awaiting the port


class PRBSConfig:
    """L23 high-speed port PRBS configuration."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.type = PP_PRBSTYPE(conn, module_id, port_id)
        """PRBS type used when in PRBS mode.
        Representation of PP_PRBSTYPE
        """

        self.tx_type = PP_TXPRBSTYPE(conn, module_id, port_id)
        """TX PRBS type used when in PRBS mode.
        Representation of PP_TXPRBSTYPE
        """

        self.rx_type = PP_RXPRBSTYPE(conn, module_id, port_id)
        """RX PRBS type used when in PRBS mode.
        Representation of PP_RXPRBSTYPE
        """


class SDEyeDiagram:
    """L23 high-speed port SerDes eye diagram."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.__serdes_index = serdes_xindex
        self.measure = PP_EYEMEASURE(conn, module_id, port_id, serdes_xindex)
        """BER eye measurement.
        Representation of PP_EYEMEASURE
        """

        self.resolution = PP_EYERESOLUTION(conn, module_id, port_id, serdes_xindex)
        """Resolution for BER eye measurement.
        Representation of PP_EYERESOLUTION
        """

        self.info = PP_EYEINFO(conn, module_id, port_id, serdes_xindex)
        """Information of BER eye measurement.
        Representation of PP_EYEINFO
        """

        self.ber = PP_EYEBER(conn, module_id, port_id, serdes_xindex)
        """BER estimation of an eye diagram.
        Representation of PP_EYEBER
        """

        self.dwell_bits = PP_EYEDWELLBITS(conn, module_id, port_id, serdes_xindex)
        """Dwell bits for an eye capture.
        Representation of PP_EYEDWELLBITS
        """

    def __await__(self):
        return self._setup().__await__()

    async def _setup(self) -> Self:
        resolution = await self.resolution.get()
        self.read_column = tuple(
            PP_EYEREAD(
                self.__conn,
                self.__module_id,
                self.__port_id,
                self.__serdes_index,
                _colum_xindex=x
            )
            for x in range(resolution.x_resolution)
        )
        return self


class SDPhy:
    """L23 high-speed port SerDes PHY configuration and status."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.tx_equalizer = PP_PHYTXEQ(conn, module_id, port_id, serdes_xindex)
        """Equalizer settings of the on-board PHY in the TX direction.
        Representation of PP_PHYTXEQ
        """

        self.rx_equalizer = PP_PHYRXEQ(conn, module_id, port_id, serdes_xindex)
        """Equalizer settings of the on-board PHY in the RX direction.
        Representation of PP_PHYRXEQ
        """

        self.retune = PP_PHYRETUNE(conn, module_id, port_id, serdes_xindex)
        """Retuning of the PHY.
        Representation of PP_PHYRETUNE
        """

        self.autotune = PP_PHYAUTOTUNE(conn, module_id, port_id, serdes_xindex)
        """Autotune of the PHY.
        Representation of PP_PHYAUTOTUNE
        """


class SDPma:
    """L23 high-speed port PMA"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.precoding_config = PP_PRECODING(conn, module_id, port_id, serdes_xindex)
        """GET/SET Pre-Coding Configurations. (only for Freya)

        :type: PP_PRECODING
        """

        self.precoding_status = PP_PRECODINGSTATUS(conn, module_id, port_id, serdes_xindex)
        """GET/SET Pre-Coding Configurations. (only for Freya)

        :type: PP_PRECODING
        """

        self.graycoding = PP_GRAYCODING(conn, module_id, port_id, serdes_xindex)
        """GET/SET Gray-Coding Configurations. (only for Freya)

        :type: PP_GRAYCODING
        """


class Prbs:
    """L23 high-speed port SerDes PRBS configuration and status."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.tx_config = PP_TXPRBSCONFIG(conn, module_id, port_id, serdes_xindex)
        """TX PRBS configuration of a SerDes.
        Representation of PP_TXPRBSCONFIG
        """

        self.status = PP_RXPRBSSTATUS(conn, module_id, port_id, serdes_xindex)
        """RX PRBS status on a SerDes
        Representation of PP_RXPRBSSTATUS
        """

        self.config = PRBSConfig(conn, module_id, port_id)
        """
        L23 high-speed port PRBS configuration.
        """


class SerDes:
    """L23 high-speed port SerDes configuration and status."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.prbs = Prbs(conn, module_id, port_id, serdes_xindex)
        """PRBS configuration"""

        self.phy = SDPhy(conn, module_id, port_id, serdes_xindex)
        """PHY configuration"""

        self.eye_diagram = SDEyeDiagram(conn, module_id, port_id, serdes_xindex)
        """Eye diagram"""

        self.pma = SDPma(conn, module_id, port_id, serdes_xindex)
        """PMA layer"""

    def __await__(self):
        return self._setup().__await__()

    async def _setup(self) -> Self:
        await self.eye_diagram
        return self
